package PureMessage::Manager::Policy;
use strict;
use Carp qw(croak);
use PureMessage::Config qw(expand_config_filename expand_filename);
use ActiveState::Config qw(path_node path_nodes path_children path_value path_values);
use PureMessage::Policy::Config;
use base 'Exporter';
our @EXPORT_OK = qw(MIDDLE_COL RIGHT_COL add_cmd_before);

sub MIDDLE_COL () {-1}
sub RIGHT_COL () {-2}

sub new {
    my $proto = shift;
    my $CLASS = ref($proto) || $proto;
    my $self = { @_ };
    bless $self, $CLASS;    
    $self->{conf_file} ||= expand_filename('etc/data/policy/base.conf');
    $self->{conf} = PureMessage::Policy::Config->new();
    $self->{conf}->load_plugin($self->{conf_file}, usecache=>1);
    return $self;
}

sub rule {
    my $self = shift;
    return PureMessage::Manager::Policy::Rule->new(@_, policy=>$self);
}

sub test_list {
    my $self = shift;    
    return $self->_get_list('test');
}

sub action_list {
    my $self = shift;
    my @list;
    return $self->_get_list('action');
}

sub _get_list {
    my $self = shift;
    my $type = shift;
    
    my $key = "${type}_list";
    if (!$self->{$key}) {
        my @list;
        foreach my $name ($self->{conf}->nodes_of_type($type)) {
            my $node = $self->{conf}->get_node($name);
            my $visible = path_value($node,'visible');
            foreach my $shortcut ($self->{conf}->nodes_of_type('shortcut')) {
                my $node = $self->{conf}->get_node($shortcut);
                if (path_value($node,'name') eq $name) {
                    push @list, $shortcut;
                }
            }
            push @list, $name unless defined $visible and $visible==0;
        }
        $self->{$key} = $self->_command_factory(@list);
    }

    return @{$self->{$key}};
}

# class method
sub add_cmd_before {
    my $cmd = shift;
    my $what = shift;
    my @actions = @_;
    
    # look for things to insert before
    my $found;
    for (my $i = 0; $i < @actions; $i++) {
        $found = $i, last if grep $actions[$i]->name eq $_, @$what;
    }
    if (defined $found) {
        # if there is one
        if ($found==0) {
            # prepend
            unshift @actions, $cmd;
        }
        else {
            # insert right before
            @actions = (@actions[0..$found-1], $cmd, @actions[$found..$#actions]);
        }
    }
    else {
        # else put it on the end
        push @actions, $cmd;
    }
    return @actions;
}

# debug output - $self->_d("");
sub _d {
    my $self = shift;
    print @_ if $self->{debug};
}

sub _command_factory {
    my $self = shift;
    my @args = @_;
    my @cmds;
    foreach my $arg (@args) {
        push @cmds, PureMessage::Manager::Policy::Command->new($arg, policy=>$self);
    }
    return \@cmds;
}

1;

package PureMessage::Manager::Policy::Entity;
use strict;
use Carp qw(croak);

sub _get_name {
    my $self = shift;
    my $in_comment = shift;
    
    my $name;
    my @out_comments;
    foreach my $comment (split "\n", $in_comment) {
        $comment =~ /^attr[ \t]*NAME[ \t]*=[ \t]*(.*)$/;
        if ($1) {
            $name = $1;
        }
        else {
            push @out_comments, $comment;
        }
    }
    return wantarray ? ($name,\@out_comments) : $name;
}

sub _put_name {
    my $self = shift;
    my $name = shift;
    my $in_comment = shift;

    push @$in_comment, "attr NAME=$name";
    
    return join("\n", @{$self->{comments}});
}

sub _flatten {
    my $self = shift;
    map { ref($_) eq 'ARRAY' ? @$_ : $_ } @_;
}

1;

package PureMessage::Manager::Policy::Rule;
use strict;
use base 'PureMessage::Manager::Policy::Entity';
use Carp qw(croak);
use PureMessage::Sieve::Nodes qw(new_cond new_test new_testlist);

sub new {
    my $proto = shift;
    my $CLASS = ref($proto) || $proto;
    my $arg = shift;
    my %opts = @_;
    
    my $self = \%opts;
    bless $self, $CLASS;
    
    if (ref($arg) eq 'PureMessage::Sieve::Cond') {
        $self->put_node($arg);
    }
    elsif (! ref($arg)) {
        $self->name($arg);
    }
    else {
        croak "$CLASS must either be initialized with Sieve 'Cond' node or a name";
    }
    
    return $self;
}

sub put_node {
    my $self = shift;
    my $cond = shift;
    
    $self->{node} = $cond;
    $self->{anyof} = 1;
    
    my ($name, $comments) = $self->_get_name($cond->comment);

    $self->{comments} = $comments;
    $self->{name} = $name || "Untitled rule";
    
    my $subtest;
    if (my $test = $self->{node}->test()) {
        if ($test->name eq 'allof') {
            # all of...
            $self->{anyof} = 0;
            $self->{tests} = $self->{policy}->_command_factory(($test->args)[0]->tests);
        }
        elsif ($test->name eq 'anyof') {
            # any of...
            $self->{anyof} = 1;
            $self->{tests} = $self->{policy}->_command_factory(($test->args)[0]->tests);
        }
        elsif ($test->name eq 'not') {
            $subtest = ($test->args)[0];
            if ($subtest->name eq 'allof') {
                # not all of...
                $self->{anyof} = 3;
                $self->{tests} = $self->{policy}->_command_factory(($subtest->args)[0]->tests);
            }
            elsif ($subtest->name eq 'anyof') {
                # none of...
                $self->{anyof} = 2;
                $self->{tests} = $self->{policy}->_command_factory(($subtest->args)[0]->tests);
            }
            else {
                # none of with single subtest
                $self->{anyof} = 2;
                $self->{tests} = $self->{policy}->_command_factory($subtest);
                $self->{anyof} = 1 if $self->{tests}->[0]->negate;
            }
        }
        else {
            $self->{tests} = $self->{policy}->_command_factory($test);
        }
        
        if (scalar @{$self->{tests}} > 1) {
           my $i = 0;
           for (my $i = 0; $i < scalar @{$self->{tests}}; $i++) {
                if ($self->{tests}->[$i]->name eq 'not') {
                    my $negated_test = ($self->{tests}->[$i]->{node}->args)[0];
                    my $o = $self->{policy}->_command_factory($negated_test);
                    croak "cannot negate test ".$o->[0]->name
                        unless $o->[0]->negate;
                    splice @{$self->{tests}}, $i, 1, $o->[0];
                }
            }
        }
 
    }

    if (my @actions = grep ref($_) eq "PureMessage::Sieve::Action", $self->{node}->commands()) {
        $self->{actions} = $self->{policy}->_command_factory(@actions);
    }
}

sub name {
    my $self = shift;
    my $name = shift;
    return $self->{name} unless defined $name;
    $self->{name} = $name;
}

sub tests {
    return @{shift->{tests} || []};
}

sub test {
    return shift->{tests}->[shift];
}

sub action {
    return shift->{actions}->[shift];
}

sub actions {
    return @{shift->{actions} || []};
}

sub anyof {
    my $self = shift;
    my $anyof = shift;
    return $self->{anyof} unless defined $anyof;
    $self->{anyof} = $anyof;
}

sub raw_sieve {
    return "raw sieve not implemented yet";
}

sub add_test {
    my $self=shift;
    my $cmd = PureMessage::Manager::Policy::Command->new(@_, policy=>$self->{policy});
    push @{$self->{tests}}, $cmd;
    return $cmd;
}

sub add_action {
    my $self=shift;
    my $cmd = PureMessage::Manager::Policy::Command->new(@_, policy=>$self->{policy});
    
    # put the new action before the first stop command, if there is one
    @{$self->{actions}} = PureMessage::Manager::Policy::add_cmd_before(
                    $cmd,
                    ['stop'],
                    @{$self->{actions}}
                );
    return $cmd;   
}

sub get_node {
    my $self = shift;
    
    my $test;
    my @test_nodes;
    
    foreach ($self->tests) {
        push @test_nodes, $_->get_node;
    }
    
    if (scalar @test_nodes) {
        my $subtest;
        if (scalar @test_nodes == 1) {
            $subtest = $test_nodes[0];
        }
        else {
            if ($self->{anyof} == 0 or $self->{anyof} == 3) {
                $subtest = new_test('allof', new_testlist(@test_nodes));
            }
            elsif ($self->{anyof} == 1 or $self->{anyof} == 2) {
                $subtest = new_test('anyof', new_testlist(@test_nodes));
            }
            
        }
        
        if ($self->{anyof} == 0 or $self->{anyof} == 1) {
            $test = $subtest;
        }
        elsif ($self->{anyof} == 2 or $self->{anyof} == 3) {
            $test = new_test('not',$subtest);
        }
        else {
            croak "unknown anyof value '$self->{anyof}'";
        }
    }
    
    my @commands;
    my $i = 0;
    foreach my $cmd ($self->{node}->commands) {
        if (ref($cmd) eq 'PureMessage::Sieve::If') {
            push @commands, $cmd;
        }
        else {
            push @commands, $self->{actions}->[$i]->get_node if $self->{actions}->[$i];
            $i++;
        }
    }
    if ($self->{actions}) {
        for (; $i < scalar @{$self->{actions}}; $i++) {
            push @commands, $self->{actions}->[$i]->get_node;
        }
    }
    
    return new_cond($test,
                        $self->_put_name($self->{name}, $self->{comments}),
                        @commands);
}

sub save_node {
    my $self = shift;

    $self->{pos} or croak "no position saved!";

    my @conds = $self->{pos}->[0]->args;

    my $node =  $self->get_node;

    $conds[$self->{pos}->[1]] = $node;
    
    $self->{pos}->[0]->args(@conds);
    
    return $node;
}

sub delete_node {
    my $self = shift;

    $self->{pos} or croak "no position saved!";
    
    my @conds = $self->{pos}->[0]->args;

    my $node = splice @conds, $self->{pos}->[1], 1;
    
    $self->{pos}->[0]->args(@conds);

    return $node;
}

1;

package PureMessage::Manager::Policy::Command;
use strict;
use base 'PureMessage::Manager::Policy::Entity';
use Carp qw(croak);
use ActiveState::Config qw(path_node path_nodes path_children path_value path_values);
use PureMessage::Sieve::Nodes qw(new_test new_action new_string new_stringlist new_string new_tag new_number);

sub new {
    my $proto = shift;
    my $CLASS = ref($proto) || $proto;
    my $arg = shift;
    my %opts = @_;
    my $self = bless \%opts, $CLASS;

    if (ref($arg) =~ /^PureMessage::Sieve::(?:Test|Action)$/) {
        $self->put_node($arg);
    }
    elsif (! ref($arg)) {
        $self->name($arg);
    }
    else {
        croak "$CLASS must either be initialized with Sieve command node or a name";
    }
    return $self;
}

sub put_node {
    my $self = shift;
    my $node = shift;
    my $sieve_args =
        $self->{policy}->{conf}->validate_node(
                                    $node,
                                    node_args => {
                                        no_lists => 1,
                                        no_templates => 1,
                                    }
        );
    my $shortcut;
    SHORTCUT: foreach my $sc_name ($self->{policy}->{conf}->nodes_of_type('shortcut')) {
        my $sc = $self->{policy}->{conf}->get_node($sc_name);
        if (path_value($sc,'name') eq $node->name) {
            foreach my $arg ($sc->children) {
                my $arg_type = $self->_arg_type($arg);
                next unless $arg_type =~ /^(?:param|tag_set)$/;
                my $value = $sieve_args->arg_by_name($arg->name);
                next SHORTCUT unless defined $value;
                next SHORTCUT unless defined (my $pvalue = path_value($arg,'value'));
                if ($arg_type eq 'param') {
                    my $conf = $self->{policy}->{conf}->get_node($node->name);
                    if (path_value($conf,'param#'.$arg->name.'.type') eq 'number') {
                        next SHORTCUT unless $pvalue == $value;
                    }
                    else {
                        my @value = sort $self->_flatten($value);
                        my @pvalue = sort $self->_flatten($pvalue);
                        next SHORTCUT if !@value or (scalar @value != scalar @pvalue);
                        for (my $i = 0; $i < scalar @pvalue; $i++) {
                            next SHORTCUT unless lc($pvalue[$i]) eq lc($value[$i]);
                        }
                    }
                }
                else {
                    next SHORTCUT unless $pvalue eq $value;
                }
                $shortcut = $sc;
                last SHORTCUT;
            }
        }
    }
    if ($shortcut) {
        $self->_name($shortcut->name);
    }
    else {
        $self->_name($node->name);
    }
    $self->{node} = $node;
    $self->{comment} = $node->comment;
    
    my $i = 0;
    foreach my $arg (@{$self->{conf_args}}) {
        my $value;
        my $type = $arg->groupclass;
        if ($type eq 'tag_set') {
            $value = $sieve_args->arg_by_name($arg->name);
        }
        elsif ($type eq 'tag') {
            my $tag = $sieve_args->tag($arg->name);
            
            if (defined $tag) {
                if ($self->{args}[$i]{type} eq 'checkbox') {
                    $value = 1;
                }
                else {
                    $value = $tag->{params}->[0]->{value} if defined $tag->{params}->[0];
                }
            }
            elsif ($self->{args}[$i]{type} eq 'checkbox') {
                $value = 0;
            }
        }
        elsif ($type eq 'param'){
            $value = $sieve_args->arg_by_name($arg->name);
        }
        else {
            croak "unexpected node type '$type'";
        }
        $self->{args}[$i]{value} = $value if defined $value;
        $i++;
    }
}

sub name {
    my $self = shift;
    my $name = shift;
    return ($self->{shortcut} ? $self->{shortcut}->name : $self->{name}) unless $name;
    delete $self->{args};
    delete $self->{node};
    delete $self->{middle_col};
    delete $self->{right_col};
    delete $self->{num_conf_args};
    delete $self->{comment};
    delete $self->{shortcut};
    delete $self->{label};
    $self->_name($name);
}

sub _name {
    my $self = shift;
    my $name = shift;
    $self->{name} = $name;
    
    my %shortcut_args;
    
    if ($self->{policy}->{conf}->is_of_type('shortcut',$self->{name})) {
        $self->{shortcut} = $self->{policy}->{conf}->get_node($self->{name});
        
        if ($self->{shortcut}) {
            foreach my $arg ($self->{shortcut}->children) {
                if ($arg->name eq 'name') {
                    $self->{name} = $arg->value;
                }
                elsif ($arg->name eq 'desc') {
                    $self->{label} = $arg->value;
                }
                elsif ($arg->name eq 'use_popup') {
                    $self->{requires_popup} = $arg->value;
                }
                elsif (grep $self->_arg_type($arg) eq $_, ('tag_set','param','tag')) {
                    $shortcut_args{$self->_arg_name($arg)} = 1;
                }
            }
        }
    }
    $self->{conf} = $self->{policy}->{conf}->get_node($self->{name});
    unless ($self->{shortcut}) {
        $self->{label} = path_value($self->{conf},'desc'); 
        $self->{requires_popup} = path_value($self->{conf},'use_popup');
    }
    die "command '$self->{name}' not found" unless defined $self->{conf};

    $self->{conf_args} = [];
    $self->_walk_conf_args(
                           sub {
                               push @{$self->{conf_args}}, shift;
                               1;
                           },
                           'tag_set','tag','param',
                          );
    $self->{num_conf_args} = scalar @{$self->{conf_args}};
    
    my $i = 0;
    foreach my $arg (@{$self->{conf_args}}) {
        my $type = $arg->groupclass;
        my $entry = {};
        $entry->{name} = $arg->name;
        my $visible = path_value($arg,'visible');
        
        if ($type eq 'tag_set') {
            $entry->{type} = 'select';
            unless (defined $visible and $visible == 0
                        or $shortcut_args{$entry->{name}}) {
                $self->{middle_col} = $i unless exists $self->{middle_col};
            }
            foreach my $node ($arg->children) {
                if ($self->_arg_type($node) eq 'tag') {
                    my $desc = path_value($node, 'desc');
                    my $alt = path_value($node, 'alt');
                    $entry->{options}{$node->name} = $desc if $desc;
                    $entry->{options}{$node->name.'_not'} = $alt if $alt;
                }
            }
            $entry->{value} = path_value($arg,'default');
        }
        elsif ($type eq 'tag') {
            $entry->{type} = 'checkbox';
            my $param_type;
            if(my ($param) = grep $self->_arg_type($_) eq 'param', $arg->children) {
                $param_type = path_value($param, 'type') or croak "argument '$entry->{name}' param '".$param->name."' has no type";
                $entry->{type} .= "_$param_type";
                $entry->{value} = path_value($param,'default');
            }
        }
        elsif ($type eq 'param') {
            my $param_type = path_value($arg, 'type') or croak "argument '$entry->{name}' has no type";
            $entry->{type} = $param_type;
            unless ($visible and $visible == 0
                        or $shortcut_args{$entry->{name}}
                        ) {
                $self->{middle_col} = $i unless exists $self->{middle_col};
                $self->{right_col} = $i unless exists $self->{right_col} or $self->{middle_col} == $i;
            }
            $entry->{value} = path_value($arg,'default');
        }
        else {
            croak "unexpected node type '$type'";
        }
        if ($self->{shortcut}) {
            my $value = path_value($self->{shortcut},"$type#$entry->{name}.value");
            $entry->{value} = $value if defined $value;
        }
        $entry->{label} = path_value($arg, 'desc');
        push @{$self->{args}}, $entry;
        $i++;
    }
    $self->{right_col} = $self->{middle_col} unless defined $self->{right_col};
}

sub label {
    my $self = shift;
    return $self->{label} || $self->{name};
}

sub requires_popup {
    my $self = shift;
    return $self->{requires_popup} || 0;
}

sub negate {
    my $self = shift;
    
    my $opt = $self->arg_selected;
    return 0 unless $opt;
    my $opts = $self->arg_options;
    return 0 unless exists $opts->{$opt.'_not'};
    
    $self->arg_selected($opt.'_not');
    return 1;
}

# this function is deprecated - use arg_value(MIDDLE_COL,...) instead. 
sub arg_selected {
    my $self = shift;
    my $selected = shift;
    my $index = shift;
    unless (exists $self->{middle_col}) {
        croak "command '$self->{name}' does not take a 'select' parameter" if defined $selected;
        return undef;
    }
    return undef unless exists $self->{middle_col};
    return $self->{args}[$self->{middle_col}]->{value} unless defined $selected;
    $self->{args}[$self->{middle_col}]->{value} = $selected;
}

sub _decode_simple_index {
    my $self = shift;
    my $index = shift;
    my $default = shift;
    my $value = shift;
    
    # XXX old behaviour - deprecated - should go away
    if (!defined $index) {
        unless (defined $default) {
            croak "tried to assign to illegal index for command '$self->{name}'" if defined $value;
            return undef;
        }
        $index = $default;
    }
    # simple 3 column interface  (left col is command name)
    elsif ($index == PureMessage::Manager::Policy::MIDDLE_COL()) {
        $index = $self->{middle_col};        
    }
    elsif ($index == PureMessage::Manager::Policy::RIGHT_COL()) {
        $index = $self->{right_col};
    }
    else {
        # complex interface, indexed parameters
    }

    croak "index '$index' out of range" if $index < 0 or $index > $self->{num_conf_args};

    return $index;
}

sub arg_options {
    my $self = shift;
    my $index = shift;
    
    $index = $self->_decode_simple_index($index, $self->{middle_col});
    return undef unless defined $index;
    return $self->{args}[$index]->{options};
}

sub arg_value {
    my $self = shift;
    my $index = shift;
    $index = $self->_decode_simple_index($index, $self->{right_col});
    return undef unless defined $index;
    if ( @_ ) {
        my $value = shift;
        $self->{args}[$index]->{value} = $value;
    }
    else {
        return $self->{args}[$index]->{value};
    }
}

sub arg_type {
    my $self = shift;
    my $index = shift;
    
    $index = $self->_decode_simple_index($index, $self->{right_col});
    return undef unless defined $index;
    return $self->{args}[$index]->{type};
}

sub arg_label {
    my $self = shift;
    my $index = shift;
    
    $index = $self->_decode_simple_index($index);
    return undef unless defined $index;
    return $self->{args}[$index]->{label};
}

sub arg_name {
    my $self = shift;
    my $index = shift;
    
    $index = $self->_decode_simple_index($index);
    return undef unless defined $index;
    return $self->{args}[$index]->{name};
}

sub args {
    my $self = shift;
    return $self->{num_conf_args};
}

sub _walk_conf_args {
    my $self = shift;
    my $cb = shift;
    my @types = @_;

    ARG: foreach my $arg_node ($self->{conf}->children) {
        my $arg_type = $self->_arg_type($arg_node);
        next unless grep $arg_type eq $_, @types;
        if ($arg_node->type eq 'option')  {
            my @arg_names = ();
            if (ref($arg_node->value) eq 'ARRAY')  {
                # AS::Config will collapse options with same names - lets expand them
                @arg_names = @{$arg_node->value}
            }
            else {
                # default case - a single option
                @arg_names = ($self->_arg_name($arg_node));
            }
            # lookup each name
            foreach my $arg_name (@arg_names) {
                my $node = $self->{policy}->{conf}->get_node($arg_name);
                croak "plugin config error: global $arg_type '$arg_name' not found" unless $node;
                last ARG unless $cb->($node);
            }
        }
        else {
            last ARG unless $cb->($arg_node);
        }
    }
}

my %conf2siv = ('test', 'PureMessage::Sieve::Test',
                    'action', 'PureMessage::Sieve::Action',
                    'tag' => 'PureMessage::Sieve::Tag',
                    'number' => 'PureMessage::Sieve::Number',
                    'string' => 'PureMessage::Sieve::String',
                    'string-list' => 'PureMessage::Sieve::Stringlist',
                    'test-list' => 'PureMessage::Sieve::Testlist',
                   );

my %conf2new = ('test', \&new_test,
                'action', \&new_action,
                'tag' => \&new_tag,
                'number' => \&new_number,
                'string' => \&new_string,
                'string-list' => \&new_stringlist,
                   );

sub _create_sieve_node {
    my $self = shift;
    my $type = shift;
    
    if ($type eq 'string') {
        my @values = $self->_flatten(@_);
        @_ = ($values[0]);
    }
    my $class = $conf2new{$type} or croak "unknown class '$type'";
    return $class->(@_);
}

sub get_node {
    my $self = shift;    
    my $negated = 0;
        
    my $i = 0;
    my @args;
    foreach my $arg (@{$self->{conf_args}}) {
        my $value = $self->{args}[$i]{value};
        my $type = $arg->groupclass;
        
        if (defined $value) {
            if ($type eq 'tag_set') {
                $negated = 1 if $value =~ s/^(.*)_not$/$1/;
                push @args, $self->_create_sieve_node('tag', ':'.$value);
            }
            elsif ($type eq 'tag') {
                if ($self->{args}[$i]{type} eq 'checkbox') {
                    push @args, $self->_create_sieve_node('tag', ':'.$self->{args}[$i]{name})
                        if $value;
                }
                else {
                    push @args, $self->_create_sieve_node('tag', ':'.$self->{args}[$i]{name});
                    my ($type) = $self->{args}[$i]{type} =~ /^checkbox_(.*)$/;
                    push @args, $self->_create_sieve_node($type, $value);
                }
            }
            else {
                push @args, $self->_create_sieve_node($self->{args}[$i]{type}, $value);
            }
        }
        $i++;
    }
        
    my $node = $self->_create_sieve_node($self->{conf}->groupclass,
                                        [$self->{name}, $self->{comment} || '', 0],
                                        @args);
    
    if ($negated) {
        $node = $self->_create_sieve_node('test', 'not', $node);
    }

    $self->{policy}->{conf}->validate_node($node);
    
    return $node;
}



sub _arg_type {
    my $self = shift;
    my $node = shift;
    return $node->type eq 'option' ? $node->name : $node->groupclass;
}

sub _arg_name {
    my $self = shift;
    my $node = shift;
    return $node->type eq 'option' ? $node->value : $node->name;
}


1;
