package PureMessage::Policy::Args;
use strict;
use Carp qw(croak);
use PureMessage::List;
use PureMessage::Map;
use PureMessage::Template qw(extract_templates expand_template);
use PureMessage::Policy::Config::Constants qw(CMD_VISITED_FIRST);
use ActiveState::UTF8 'decode_utf8';

sub new {
    my $class = shift;
    my $parsed_args = shift;
    my $self = bless $parsed_args, $class;
    
    unless (delete $self->{no_lists}) {
        $self->build_lists();
    }
    unless (delete $self->{no_templates}) {
        $self->build_templates();
    }
    return $self;
}

sub _extract_templates {
    my $param = shift;
    
    my @templates;
    if ($param->{type} and ($param->{type} eq 'string' or $param->{type} eq 'string-list')) {
        push @templates, extract_templates(@{$param->{value}});
    }
    @templates;
}


sub _all_params {
    my $self = shift;
    my @params;

    push @params, @{$self->{p}} if $self->{p};
    if ($self->{t}) {
        foreach (keys %{$self->{t}}) {
            my $tag = $self->{t}->{$_};
            push @params, @{$tag->{params}} if $tag->{params};
        }
    }
    @params;
}

sub build_templates {
    my $self = shift;

    delete $self->{tmpl};
    push @{$self->{tmpl}}, _extract_templates($_) foreach $self->_all_params; 
}

sub build_lists {
    my $self = shift;
    my $lm = shift;
    
    delete $self->{m};
    delete $self->{l};
    delete $self->{mo};
    delete $self->{lo};
    delete $self->{lists};
    
    # named lists - just store list names and a reference to the listmanager
    $self->{lm} = $lm if $lm;
    my @lists = $self->find_named_lists;
    $self->{lists} = \@lists;
    
    # anonymous lists
    $self->enumerate_params(sub{
        return if $self->{l};
        my $param = shift;
        my $mt;
        foreach (qw/is contains matches re memberof/) {
            $mt = $_, last if $self->{t}->{$_};
        }
        $mt ||= 'is';
        unless ($mt eq 'memberof') {
            my @params;
            my $cmp_tag = $self->tag('comparator');
            if (defined $cmp_tag) {
                push @params, comparator=>$cmp_tag->{params}->[0]->{value}->[0] if $cmp_tag->{params};
            }
            push @params, match_type=>$mt;
            push @params, items=>$param->{value};
            my $list = PureMessage::List->new(@params);
            $self->{l} = $list->matcher;
        }
    }, 'key-list');
    
    # maps
    $self->enumerate_params(sub{
        return if $self->{m};
        my $param = shift;
        my $map = PureMessage::Map->new($param->{value}->[0]);
        $self->{m} = $map->transformer;
        $self->{mo} = $map;
    }, 'map-name');
}

sub find_named_lists {
    my $self = shift;
        
    my @lists;
    
    # list implicitly declared in the plugin config file
    if (defined (my $use_list = $self->{use_list})) {
        push @lists, ref $use_list ? @$use_list : $use_list;
    }
    $self->enumerate_params(sub {
        my ($param, $name) = @_;
        if ($name eq 'key-list' and $self->{t}->{'memberof'}) {
            push @lists, $param->{value}->[0];
        }
        elsif ($name eq 'list-name') {
            push @lists, $param->{value}->[0];
        }
        
    },['list-name','key-list']);
    
    @lists;
}

sub templates {
    my $self = shift;
    return () unless $self->{tmpl};
    return @{$self->{tmpl}};
}

sub _no_match {
    return 0;
}

sub matcher {
    my $self = shift;
    my $key = shift;
    my $i = shift || 0;

    # return anonymous list matcher, if one exists
    return $self->{l} if $self->{l};
    
    # return named list matcher, found by ListManager 
    if ($self->{lm} and $self->{lists}) {
        my $list =  $self->{lm}->find_list($self->{lists}[$i], key=>$key);
        return $list ? $list->matcher : \&_no_match;
    }
    return undef;
}

sub find_list {
    return shift->{lm}->find_list(@_);
}

sub is_multilist {
    my $self = shift;
   
    return 0 if $self->{l};
    
    if ($self->{lm} and $self->{lists}) {
        return $self->{lm}->is_multilist($self->{lists}[0]);
    }
    return 0;
}

sub enumerate_params {
    my $self = shift;
    my $cb = shift;
    my $names = shift;
    
    my @names = ref $names ? @$names : $names;
        
    foreach my $param ($self->_all_params) {
        my $name = $param->{name};
        $cb->($param, $name) if grep $_ eq $name, @names; 
    }
}

sub transformer {
    my $self = shift;
    return $self->{m};
}

sub refresh_list {
    my $self = shift;
    
    my $rc_m;
    # only refresh the map.  lists are now refreshed by the list manager
    # undef means there is no list, or that it doesnt need refreshing
    if (exists $self->{mo}) {
        $rc_m = $self->{mo}->refresh;
        $self->{m} = $self->{mo}->transformer if $rc_m;
    }

    return $rc_m;
}

sub tags {
    my $self = shift;
    return keys %{$self->{t}};
}

sub tag {
    my $self = shift;
    my $tag = shift;
    return $self->{t}->{$tag};
}

sub params {
    my $self = shift;
    return map $_->{value}, @{$self->{p}};
}

# XXX TODO: this method should not have "set" capability.
# the parameters should be read-only after the object is created.
sub param {
    my $self = shift;
    my $i = shift;
    my $param = shift;
    unless (defined $param) {
        return undef if $i > $#{$self->{p}} or $i < 0;
        return ($self->{template}) ?
        $self->{template}->expand($self->{p}->[$i]->{value}, list=>$self->{tmpl})
        : $self->{p}->[$i]->{value};
    }
    $self->{p}->[$i]->{value} = $param;
}

sub param_decoded {
    my $self = shift;
    my $i = shift;
    return undef if $i > $#{$self->{p}} or $i < 0;
    my $p = $self->{p}->[$i]->{value};
    $p = [map decode_utf8($_), (ref $p ? @$p : $p)];
    return ($self->{template}) ?
        $self->{template}->expand($p, list=>$self->{tmpl}) : $p;
}

sub arg_by_name {
    my $self = shift;
    my $name = shift;
    my $value = shift;
    my ($i,$found) = (0,0);
    
    foreach (@{$self->{p}}) {
        next unless $_->{name};
        $found = 1, last if $_->{name} eq $name;
        $i++;
    }
    if ($found) {
        unless (defined $value) {
            return ($self->{template}) ?
            $self->{template}->expand($self->{p}->[$i]->{value}, list=>$self->{tmpl})
            : $self->{p}->[$i]->{value};
        }
        $self->{p}->[$i]->{value} = $value;
    }
    else {
        my $v;
        foreach (keys %{$self->{t}}) {
            $v = $_, last if defined $self->{t}{$_}->{tag_set} and
                $self->{t}{$_}->{tag_set}  eq $name;
        }
        return $v unless defined $value;
        $self->{t}->{$v} = $value if defined $v;
    }
}

sub type {
    my $self = shift;
    my $i = shift;
    return undef if $i > $#{$self->{p}} or $i < 0;
    return $self->{p}->[$i]->{type};
}

sub cached {
    my $self = shift;
    my $cached = shift;
    return $self->{c} unless defined $cached;
    $self->{c} = $cached;
}

sub first_time {
    my $self = shift;
    my $old = $self->{ft};
    $self->{ft} = shift if @_;
    return $old;
}

sub template_object {
    my $self = shift;
    my $old = $self->{template};
    $self->{template} = shift if @_;
    return $old;
}

sub get_multifiles {
    my ($self) = @_;
    my ($id, $charset) = (undef, undef);

    my $groupfile_tag = $self->tag('groupfile');
    my $charset_tag = $self->tag('charset');
    if($groupfile_tag) {
        $id = $groupfile_tag->{params}->[0]->{value}->[0];
        $charset = $charset_tag ? $charset_tag->{params}->[0]->{value}->[0] : undef;
    }

    return wantarray ? ($id, $charset) : $id;
}

1;
