package PureMessage::Policy::Config;
use base 'Exporter';
use strict;
use Carp qw(croak);
use ActiveState::Config qw(path_node path_nodes path_children path_value path_values);
use PureMessage::Policy::Config::Constants qw(CMD_UNVISITED CMD_VISITED_TRUE CMD_VISITED_FALSE CMD_VISITED_FIRST CMD_VISITED_REENTER);

sub new {
    my $class = shift;
    my %opts = @_;
    return bless \%opts, $class;
}

sub load_plugin {
    my $self = shift;
    my $conf = shift;
    my %opts = @_;
    
    my $usecache = delete $opts{usecache};
    my $newcache = delete $opts{newcache};

    my $p;
    
    if ($usecache) {
        my $cache = "$conf.cache";
        if (-f $cache) {
            my $cache_time = (stat $cache)[9];
            my $conf_time = (stat $conf)[9];
            if ($conf_time > $cache_time or $newcache) {
                $p = _conf_load($conf,%opts);

                # we write $$p instead of $p.
                # this means that we cache
                # the real ActiveState::Config object and
                # not the ActiveState::Config::FakeObject object
                
                _write_cache($$p,$cache);
            }
            else {
                eval{$p = _read_cache($cache)};
                if ($@) {
                    warn "Possibly corrupted $cache.  Fixing...\n";
                    $p = _conf_load($conf,%opts);
                    _write_cache($$p,$cache);
                }
            }
        }
        else {
            $p = _conf_load($conf,%opts);
            _write_cache($$p,$cache);
        }
    }
    else {
        $p = _conf_load($conf,%opts);
    }
    
    push @{$self->{plugins}}, $p;
    $self->_conf_build_lookup($p->root);
}

sub _write_cache {
    my $o = shift;
    my $f = shift;
    
    # store will warn() about trying to store coderefs, but these
    # coderefs are the validatiors - only used in validation and will never
    # be used in the cached/retrieved copy (these config files are not user-visible).
    # so here we ignore the warnings that it spits out.
    require Storable;
    
    # dont croak when trying to store coderefs, just warn
    $Storable::forgive_me = 1;
    
    require PureMessage::_sys;
    my $uniq = PureMessage::_sys::tid();
    my $store = "$f-$uniq";
    {
        local $SIG{__WARN__} = sub {warn @_ unless $_[0] =~ /^Can't store item CODE/};
        eval {Storable::store($o,$store)};
    }
    die $@ if $@;
    rename $store, $f or die $!;
}

sub _read_cache {
    my $f = shift;
    require Storable;
    my $p = Storable::retrieve($f);
    $p;
}

sub _conf_load {
    my $conf = shift;
    my $p = ActiveState::Config->new(
	strict => 0,
	default_type => ['#atom nodups','#atom','#list'],
	backup_ext => '.bak',
    );
    
    # plugin 
    $p->typedecl('plugin#*.use', '#str');
    $p->typedecl('plugin#*.visible', '#bool');
    $p->typedecl('plugin#*.desc', '#str');
    $p->default('plugin#*.visible', 'No');
    
    # tagset - global
    $p->typedecl('plugin#*.tag_set#*.desc', '#str');
    $p->typedecl('plugin#*.tag_set#*.visible', '#bool');
    $p->typedecl('plugin#*.tag_set#*.default', '#str');
    $p->typedecl('plugin#*.tag_set#*.required', '#bool');
    
    # tags - global
    $p->typedecl('plugin#*.tag#*.desc', '#str');
    $p->typedecl('plugin#*.tag#*.alt', '#str');
    $p->typedecl('plugin#*.tag#*.visible', '#str');
    
    # templates - global
    $p->typedecl('plugin#*.template_var#*.event', '#str');
    
     # tagset - local
    $p->typedecl('plugin#*.*#*.tag_set#*.desc', '#str');
    $p->typedecl('plugin#*.*#*.tag_set#*.visible', '#bool');
    $p->typedecl('plugin#*.*#*.tag_set#*.default', '#str');
    $p->typedecl('plugin#*.*#*.tag_set#*.required', '#bool');   
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.desc', '#str');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.alt', '#str');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.visible', '#str');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.use_list', '#list');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.param#*.type', '#str');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.param#*.value', '#list');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.param#*.default', '#str');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.param#*.required', '#bool');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.param#*.units', '#str');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.param#*.desc', '#str');
    $p->typedecl('plugin#*.*#*.tag_set#*.tag#*.param#*.visible', '#bool');
    $p->default('plugin#*.*#*.tag_set#*.tag#*.param#*.visible', 'Yes');
    
    # tags - local
    $p->typedecl('plugin#*.*#*.tag#*.desc', '#str');
    $p->typedecl('plugin#*.*#*.tag#*.alt', '#str');
    $p->typedecl('plugin#*.*#*.tag#*.visible', '#str');    
    $p->typedecl('plugin#*.*#*.tag#*.param#*.type', '#str');
    $p->typedecl('plugin#*.*#*.tag#*.param#*.value', '#list');
    $p->typedecl('plugin#*.*#*.tag#*.param#*.default', '#str');
    $p->typedecl('plugin#*.*#*.tag#*.param#*.required', '#bool');
    $p->typedecl('plugin#*.*#*.tag#*.param#*.units', '#str');
    $p->typedecl('plugin#*.*#*.tag#*.param#*.desc', '#str');
    $p->typedecl('plugin#*.*#*.tag#*.param#*.visible', '#bool');
    $p->default('plugin#*.*#*.tag#*.param#*.visible', 'Yes');
    
    # param - local and global tags
    $p->typedecl('plugin#*.*#*.param#*.type', '#str');
    $p->typedecl('plugin#*.*#*.param#*.value', '#list');
    $p->typedecl('plugin#*.*#*.param#*.default', '#str');
    $p->typedecl('plugin#*.*#*.param#*.required', '#bool');
    $p->typedecl('plugin#*.*#*.param#*.units', '#str');
    $p->typedecl('plugin#*.*#*.param#*.desc', '#str');
    $p->typedecl('plugin#*.*#*.param#*.visible', '#bool');
    
    # test
    $p->typedecl('plugin#*.test#*.visible', '#bool');
    $p->typedecl('plugin#*.test#*.event', '#str');
    $p->typedecl('plugin#*.test#*.use_popup', '#bool');

    # action
    $p->typedecl('plugin#*.action#*.visible', '#bool');
    $p->typedecl('plugin#*.action#*.event', '#str');
    
    # templates - local
    $p->typedecl('plugin#*.*#*.template_var#*.local', '#bool');
    
    # shortcuts
    $p->typedecl('plugin#*.shortcut.test', '#str');
    $p->typedecl('plugin#*.shortcut.action', '#str');
    $p->typedecl('plugin#*.shortcut.desc', '#str');
    $p->typedecl('plugin#*.shortcut.param#*.value', '#list');
    $p->typedecl('plugin#*.shortcut.tag#*.param#*.value', '#str');
    $p->typedecl('plugin#*.shortcut.tag_set#*.value', '#str');
    $p->typedecl('plugin#*.shortcut.tag_set#*.tag#*.param#*.value', '#str');

    $p->parse_file($conf, @_);
    $p;
}

# build/add to a dictionary and lookup table for conf nodes 
#
# croak an error on duplicate entries

sub _conf_build_lookup {
    my $self = shift;
    my $root_node = shift;
    foreach my $plugin_node ($root_node->children) {
        next if $plugin_node->type eq 'placeholder';
        my $use_statement =  path_value($plugin_node,"use");
        my $namespace =  path_value($plugin_node,"namespace");
        foreach my $node ($plugin_node->children) {
            next if $node->type eq 'placeholder';
            if (_arg_type($node) =~ /^(?:test|action|template_var|tag|tag_set|shortcut)$/) {
                croak ("plugin config error: duplicate declarations of '".
                        $node->name."'")
                    if exists $self->{dict}->{$node->name};
                    
                #XXX check for duplicate tags in commands that have multiple
                #XXX tag_sets as arguments
                
            $self->{dict}->{$node->name} = $node;
            $self->{types}->{$node->groupclass}{$node->name} = 1;        
            $self->{modules}{$use_statement}{$node->name} = 1;
            $self->{namespaces}{$namespace}{$node->name} = 1 if $namespace;
            }      
        }
    }
}

sub nodes {
    my $self = shift;    
    my @a = ();
    @a = keys %{$self->{dict}} if $self->{dict};
    @a;
}

sub get_node {
    my $self = shift;
    my $key = shift;
    return $self->{dict}{$key};
}

sub types {
    my $self = shift;    
    my @a = ();
    @a = keys %{$self->{types}} if $self->{types};
    @a;
}

sub nodes_of_type {
    my $self = shift;
    my $key = shift;
    my @a = ();
    @a = keys %{$self->{types}->{$key}} if $self->{types}->{$key};
    @a;
}

sub is_of_type {
    my $self = shift;
    my $type = shift;
    my $key = shift;
    return exists $self->{types}{$type}{$key};
}

sub plugins {
    my $self = shift;
    my @a = ();
    @a = @{$self->{plugins}} if $self->{plugins};
    @a;
}

sub modules {
    my $self = shift;
    my @a = ();
    @a = keys %{$self->{modules}} if $self->{modules};
    @a;
}

sub nodes_of_module {
    my $self = shift;
    my $key = shift;
    my @a = ();
    @a = keys %{$self->{modules}->{$key}} if $self->{modules}->{$key};
    @a; 
}

sub is_provided_by_module {
    my $self = shift;
    my $module = shift;
    my $key = shift;
    return exists $self->{modules}->{$module}->{$key};
}

sub namespaces {
    my $self = shift;
    my @a = ();
    @a = keys %{$self->{namespaces}} if $self->{namespaces};
    @a;
}

sub is_in_namespace {
    my $self = shift;
    my $ns = shift;
    my $key = shift;
    return exists $self->{namespaces}{$ns}{$key};
}

sub is_provided {
    my $self = shift;
    my $ns = shift;
    return exists $self->{namespaces}->{$ns};
}

my %sieve2conf = (
                    'PureMessage::Sieve::Test' => 'test',
                    'PureMessage::Sieve::Action' => 'action',
                    'PureMessage::Sieve::Tag' => 'tag',
                    'PureMessage::Sieve::Number' => 'number',
                    'PureMessage::Sieve::String' => 'string',
                    'PureMessage::Sieve::Stringlist' => 'string-list',
                    'PureMessage::Sieve::Testlist' => 'test-list',
                   );

sub _arg_type {
    my $node = shift;
    return $node->type eq 'option' ? $node->name : $node->groupclass;
}

sub _arg_name {
    my $node = shift;
    return $node->type eq 'option' ? $node->value : $node->name;
}

sub validate_node {
    my $self = shift;
    my $node = shift;
    my %opts = @_;
    
    my %arg_opts;
    %arg_opts = %{$opts{node_args}} if $opts{node_args};
        
    my $filename;
    $filename = $opts{filename} || 'policy script';
    
    my $cmd = $node->name;
    my $line = $node->line;
    my $cmdtype = $sieve2conf{ref $node};
            
    # make sure its a valid command
    croak "$filename($line): unknown $cmdtype $cmd\n"
        unless $self->is_of_type($cmdtype,$cmd);
    
    # get all the relevent conf argument nodes
    my @arg_nodes = ();
    my $cmd_node = $self->get_node($cmd);
    foreach my $arg_node ($cmd_node->children) {
        next if $arg_node->type eq 'placeholder';
        my $arg_type = _arg_type($arg_node);
        next if $arg_type !~ /^(?:tag|tag_set|param)$/;
        
        # plain options in command blocks are refs to globals - look them up
        if ($arg_node->type eq 'option')  {
            my @arg_names = ();
            if (ref($arg_node->value) eq 'ARRAY')  {
                # AS::Config will collapse options with same names - lets expand them
                @arg_names = @{$arg_node->value}
            }
            else {
                # default case - a single option
                @arg_names = (_arg_name($arg_node));
            }
            # lookup each name
            foreach my $arg_name (@arg_names) {
                croak "plugin config error: global $arg_type '$arg_name' not found"
                unless $self->is_of_type($arg_type,$arg_name);
                
                # add to the list of arg nodes
                push @arg_nodes, $self->get_node($arg_name);
            }
        }
        else {
            # its a group - i.e. locally defined - just add it
            push @arg_nodes, $arg_node;
        }
        
    }            
    
    # foreach argument to the sieve command
    my %ok_nodes = ();
    my @tag_params = ();
    my $current_tag;
    my $parsed_args = {};
    my $list_name;
    my $key_list;
    my $i = 1;
    foreach my $arg ($node->args) {
        croak "$filename($line): $cmd($i): unexpected scalar $arg\n"
            if ! ref($arg);
        croak "$filename($line): $cmd($i): ".ref($arg)." is not a valid Sieve node type\n"
            unless exists $sieve2conf{ref $arg};
        
        # type of the sieve argument
        my $type = $sieve2conf{ref $arg};
        
        my $value;
        if ($type eq 'string' or $type eq 'number' or $type eq 'tag') {
            $value = $$arg;
        }
        else {
            $value = $arg;
        }
        
        # if the sieve argument is a tag
        if ($type eq 'tag') {
            $value =~ s/^://;
            
            # if there are supposed to be parameters for the last tag,
            # we shouldnt be looking at another tag now
            if (@tag_params) {
                my $arg_type = _arg_type($tag_params[0]);
                croak "$filename($line): $cmd($i): found a tag(':$value'), expected a $arg_type\n"; 
            }
            # look through the conf argument list for a match
            # tags can appear in any order
            my $tag_match;
            my $tag_set;
            TAG: foreach my $arg_node (@arg_nodes) {
                my $arg_type = _arg_type($arg_node);
                my $arg_name = _arg_name($arg_node);
                if ($arg_type =~ /^tag(?:_set)?$/)  {                            
                    # if its a tag_set
                    if ($arg_type eq 'tag_set')  {
                        # try to match a tag in it
                        $tag_match = path_node($arg_node,"tag#$value");
                        $tag_set = $arg_node->name if $tag_match;
                    }
                    else {
                        # match the tag itself
                        $tag_match = $arg_node if $arg_name eq $value;
                    }
                    $ok_nodes{$arg_name}=1, last TAG if $tag_match;
                }
                
            }
            # make sure we had a match, if not this tag isnt valid
            croak "$filename($line): $cmd($i): invalid tag ':$value'\n" unless $tag_match;
            
            # if this tag expects subparameters, make sure we look for them next
            foreach my $child ($tag_match->children) {
                push @tag_params, $child if _arg_type($child) eq 'param';
            }
            $current_tag = $value;
            $parsed_args->{t}->{$value} = {
                                tag_set => $tag_set,
                                params => [],
                                index => $i - 1,
                            };
        }
        else {
            # this is a non-tag sieve parameter - order is significant
            
            # if there are any tag subparams on the stack, compare against the next one
            my $arg_node = shift @tag_params;
                        
            # else compare against regular params
            unless ($arg_node) {
                # reset current_tag
                $current_tag = "";
                
                # seek to the next expected param
                foreach my $arg (@arg_nodes) {
                    $arg_node = $arg, last
                        unless _arg_type($arg) =~ /^tag(?:_set)?$/
                        or exists $ok_nodes{_arg_name($arg)};
                }
            }
            
            croak "$filename($line): $cmd($i): too many parameters for this command\n"
                unless $arg_node;
            
            # do the comparing
            my $arg_name = _arg_name($arg_node);
            my $param_type = path_value($arg_node,'type');
            my $ok = 0;
            $ok = 1 if $type eq 'string' and $param_type eq 'string-list'; # string = string-list with one element
            $ok = 1 if $type eq 'test' and $param_type eq 'test-list'; # test = test-list with one element
            $ok = 1 if $param_type eq $type;
            croak "$filename($line): $cmd($i): found a $type, expected a $param_type\n"
                unless $ok;
            $ok_nodes{$arg_name} = 1;
            
            
            # de-objectify string-lists
            if ($type eq 'string-list') {
                my $tmp_value = [];
                foreach my $s (@$value) {
                    push @$tmp_value, $$s;
                }
                $value = $tmp_value;
            }
            
            # add to parsed_args if its a regular arg type
            if ($type eq 'string' or $type eq 'number' or $type eq 'string-list') {
                $value = [$value] if $type eq 'string'; # string => string-list - makes execution simpler
                my $entry = {
                    name=>$arg_name,
                    value=>$value,
                    type=>$type,
                    index => $i-1,
                };
                
                if ($current_tag) {
                    # this is a tagged argument
                    push @{$parsed_args->{t}->{$current_tag}->{params}}, $entry;
                }
                else {
                    push @{$parsed_args->{p}}, $entry;
                }
            }

            #XXX todo: check value against allowable values
        }
        $i++;
    }
    
    # check to see if any required args are missing
    # XXX for now we'll say that all "param" arguments are required
    foreach my $arg_node (@arg_nodes) {
        my $arg_name = _arg_name($arg_node);
        croak "$filename($line): $cmd(): missing required argument '$arg_name'\n"
            if (path_value($arg_node,'required')
                or _arg_type($arg_node) eq 'param'
                )
                and ! path_value($arg_node,'default')
                and ! $ok_nodes{$arg_name};
    }
    
    # return args object
    $parsed_args->{v} = CMD_UNVISITED;
    
    # check for list implicitly declared in plugin config file
    my $use_list = path_value($cmd_node, "use_list");
    $parsed_args->{use_list} = $use_list if defined $use_list;
    
    if (%arg_opts) {
        foreach my $key (%arg_opts) {
            $parsed_args->{$key} = $arg_opts{$key};
        }
    }
    
    require PureMessage::Policy::Args;
    return 
        PureMessage::Policy::Args->new($parsed_args, %arg_opts) ;
    
    #XXX the parsed args make the regular arg objects redundant.
    #XXX perhaps we should remove them to save space
}

1;
