package PureMessage::Policy::Engine;

use strict;
use Carp qw(croak);

use PureMessage::Version qw(Version $BUILD);
use PureMessage::Policy::Events qw(event_passed);
use PureMessage::Policy::Template;
use PureMessage::Policy::Logger;
use PureMessage::Policy::Script;
use PureMessage::_sys;
use Mail::Address;
use PureMessage::Map;
use PureMessage::Preferences;
use PureMessage::Group::CDBMap;

our $VERSION = "5.601";

sub _croak {
    $| = 1; $| = 0; # flush stdout first
    croak @_;
}

sub new {
    my $class = shift;
    my %opts = @_;
    my $self = bless (\%opts, $class);

    my $script = delete $self->{script};
    my $plugins = delete $self->{plugins};
    my $init = delete $self->{init};
    my $newcache = delete $self->{newcache};
    my $logger = delete $self->{logger};
    my $message_exception_action = delete $self->{message_exception_action};

    $self->{pmxpe_list_refresh_interval} ||= 60; #default - 1 minute
    $self->{pmxpe_last_refresh} = time;
    $self->{pmxpe_pc} = -1;
    $self->{pmxpe_logger} = $logger || PureMessage::Policy::Logger->new();
    $self->{pmxpe_newcache} = $newcache;
    $self->{pmxpe_message_exception_action} = 
        $message_exception_action || 'quarantine';                                              

    if ($script) {
        $self->script($script);
    }
    if ($plugins) {
        $self->add_plugin(@$plugins);
    }
    if ($init) {
        $self->init();
    }
    $self;
}

sub log {
    shift->{pmxpe_logger}->log(@_);
}

sub logger {
    my $self = shift;
    my $old = $self->{pmxpe_logger};
    $self->{pmxpe_logger} = shift if @_;
    return $old;
}

sub script {
    my $self = shift;
    my $old = $self->{pmxpe_script_file};
    $self->{pmxpe_script_file} = shift if @_;
    return $old;
}

sub add_plugin {
    my $self = shift;
    my @conf_filenames = @_;

    foreach (@conf_filenames) {
        push @{$self->{pmxpe_plugin_files}}, $_ unless $self->{pmxpe_init};
    }
}

sub init {
    my $self = shift;

    return if $self->{pmxpe_init};

    $self->{pmxpe_script} = PureMessage::Policy::Script->new(
                            $self->{pmxpe_script_file},
                            $self->{pmxpe_plugin_files},
                            newcache => $self->{pmxpe_newcache}
                        );

    my @instrs = $self->{pmxpe_script}->instruction_list;

    my ($i, @per_recipient) = 0;
    foreach my $instr (@instrs) {
        push @per_recipient, $i if $instr->{per_recipient};
        $i++;
    }

    $self->{pmxpe_instructions} = \@instrs;

    $self->{pmxpe_per_recipient_instrs} = \@per_recipient;

    my %cmds = $self->{pmxpe_script}->command_hash;

    $self->{pmxpe_commands} = \%cmds;

    $self->_load_plugins();

    $self->_load_templates();

    $self->_build_recipient_map();

    $self->_load_trusted_relays();

    $self->_load_internal_hosts();

    $self->_build_group_map();

    $self->{pmxpe_init} = 1;
}

my %internal_commands = (
                        anyof => 1,
                        allof => 1,
                        stop => 1,
                        not => 1,
                       );
my %builtin_commands = (
			PMX_VERSION => 1,
		       );

sub _load_plugins {
    my $self = shift;
    my @loaded_plugins;
    my @failed_plugins;
    my $commands = $self->{pmxpe_commands};

    foreach my $cmd (keys %$commands) {
        if (exists $internal_commands{$cmd}) {
            # delete the commands that are used internally by the engine
            delete $commands->{$cmd};
        }
        else {
            # set all commands to 0
            $commands->{$cmd} = 0;
        }
    }

    # Each module can declare its version. This will then be exposed via the
    # %%PMX_VERSION%% template variable.
    my @ver;

    foreach my $class ($self->{pmxpe_script}->required_modules) {
        # load base module - no license check
        if ($class eq 'PureMessage::Policy::Base') {
            eval "require $class";
            croak($@) if $@;
        }

        # check the license and load
        else {
            eval { PureMessage::_sys::load_module($class) };

            if ($@) {
                push @failed_plugins, $class;
                if ($@ =~ /license check failed/) {
                    $self->log('ERR', "POLICY ERROR: No license for $class found\n");
                }
                else {
                    (my $msg = $@) =~ s/Compilation failed .*?\n//m;
                    $self->log('ERR', "POLICY ERROR: Error loading $class: $msg\n");
                }
                next;
            }
        }

        push @loaded_plugins, $class;

        # get the coderef lookup
        my %lookup = $class->commands($commands, $self->{pmxpe_logger});

        push @ver, delete $lookup{__version__} if exists $lookup{__version__};

        # fill our master lookup, checking for duplicates
        foreach my $cmd (keys %lookup) {
            _croak "POLICY ERROR: plugin $class provided a duplicate of command $cmd\n"
                if $commands->{$cmd};

            # add command to lookup, but only if we wanted it in the first place
            $commands->{$cmd} = $lookup{$cmd} if exists $commands->{$cmd};
        }
    }

    # store the names of the plugins we loaded successfully
    $self->{pmxpe_loaded_plugins} = \@loaded_plugins;
    $self->{pmxpe_version} = \@ver;
    $self->log('DEBUG', "POLICY INIT: loaded plugins ".join(' ',@loaded_plugins).'.');

    # make sure all commands are accounted for
    foreach my $cmd (keys %$commands) {
	next if $builtin_commands{$cmd};
        unless (ref $commands->{$cmd} eq 'CODE') {
            my $msg = "POLICY ERROR: No plugin provided the required command '$cmd'\n";

            if (@failed_plugins) {
                $msg .= "This could be because the following plugin"
                        . (@failed_plugins == 1 ? "" : "s")
                        . " failed to load:\n";
                $msg .= "\t" . join ("\n\t", @failed_plugins) . "\n";
                $msg .= "Try running `pmx-log` for error details\n";
            }
            _croak $msg;
        }
    }
}

sub _load_templates {
    my $self = shift;

    # get templates functions needed by the policy script
    my $template_vars = $self->{pmxpe_script}->template_hash;
    foreach my $key (keys %$template_vars) {
        $template_vars->{$key} = $self->{pmxpe_commands}{$key} if $self->{pmxpe_commands}{$key};
    }

    # include any template functions that the plugins explicitly export
    foreach my $plugin (@{$self->{pmxpe_loaded_plugins}}) {
        if ($plugin->can('templates')) {
            my %templates = $plugin->templates();
            while (my ($k, $v) = each %templates) {
                $template_vars->{$k} = $v unless exists $template_vars->{$k};
            }
        }
    }

    # Provide the PMX_VERSION command
    {
	my $ver = $self->{pmxpe_version};
	# Avoid making a circular reference that includes $self
	$template_vars->{PMX_VERSION} = sub {
	    my @versions = Version() . ".$BUILD";
	    my $SEP = ": ";
	    for my $v (@$ver) {
		my @copy = @$v;
		push @versions, join($SEP, splice @copy, 0, 2) while @copy;
	    }
	    s/^PureMessage-// for @versions;
	    join(", ", @versions);
	};
    }

    # create global template expander object
    $self->{pmxpe_template} = PureMessage::Policy::Template->new($template_vars);

    # give each "args" object the global expander
    foreach my $instr (@{$self->{pmxpe_instructions}}) {
        $instr->{args}->template_object($self->{pmxpe_template}) if $instr->{args};
    }
}

sub _build_recipient_map {
    my $self = shift;
    # if the map does not exist, act the same as if it was empty
    eval {$self->{_pmxpe_rmap} = PureMessage::Map->new('recipient-aliases')};
    if ($@) {
        $self->log('DEBUG', "POLICY WARNING: Error loading recipient-aliases map: '$@'.  No recipient mapping will take place");
    }
    else {
        $self->{_pmxpe_rtx} = $self->{_pmxpe_rmap}->transformer;
    }
}

sub _load_trusted_relays {
    my $self = shift;
    # if the list does not exist, act the same as if it was empty
    eval {$self->{_pmxpe_trusted_relays} = PureMessage::List->new('trusted-relays')};
    if ($@) {
        $self->log('DEBUG', "POLICY WARNING: Error loading trusted-relays list: '$@'");
    }
}

sub _load_internal_hosts {
    my $self = shift;
    # if the list does not exist, act the same as if it was empty
    eval {$self->{_pmxpe_internal_hosts} = PureMessage::List->new('internal-hosts')};
    if ($@) {
        $self->log('DEBUG', "POLICY WARNING: Error loading internal-hosts list: '$@'");
    }
}

sub _refresh_recipient_map {
    my $self = shift;
    if ($self->{_pmxpe_rmap}) {
        $self->{_pmxpe_rmap}->refresh;
        $self->{_pmxpe_rtx} = $self->{_pmxpe_rmap}->transformer;
    }
}

sub _refresh_trusted_relays {
    my $self = shift;
    if ($self->{_pmxpe_trusted_relays}) {
        $self->{_pmxpe_trusted_relays}->refresh;
    }
}

sub _refresh_internal_hosts {
    my $self = shift;
    if ($self->{_pmxpe_internal_hosts}) {
        $self->{_pmxpe_internal_hosts}->refresh;
    }
}

sub get_relay_matchers {
    my $self = shift;
    my %matchers = ();
    $matchers{'trusted-relays'} = $self->{_pmxpe_trusted_relays}->matcher()
        if ($self->{_pmxpe_trusted_relays});
    $matchers{'internal-hosts'} = $self->{_pmxpe_internal_hosts}->matcher()
        if ($self->{_pmxpe_internal_hosts});
    return \%matchers;
}

sub map_recipients {
    my $self = shift;
    my $msg = shift;
    my @mrecips;

    local $Mail::Address::VERBOSE = 0;
    foreach my $recip (@_) {
        my ($clean) = map lc $_->address, Mail::Address->parse($recip);
        my $mrecip;
        if ($self->{_pmxpe_rmap}) {
            (my $nonplussed_addr = $clean) =~ s/\+[^@]*//;
            $mrecip = lc $self->{_pmxpe_rtx}->($nonplussed_addr);

            # Keep the original unmodified address if no mapping took place
            if ($mrecip eq $nonplussed_addr) {
                $mrecip = $clean;
            }
            else {
                $msg->log('DEBUG', "POLICY ENGINE: mapped recipient '$clean' to '$mrecip'");
            }
        }
        else {
            $mrecip = $clean;
        }
        push @mrecips, $mrecip;
    }
    return wantarray ? @mrecips : $mrecips[0];
}

sub get_direction {
    my ($self, $msg) = @_;
    return 'Inbound' unless $msg;

    if($self->{_pmxpe_internal_hosts}) {
        # Copied from pmx_relay()
        my $relay = $msg->minfo_value("relay");
        if($relay) {
            my @relay = split / /, $relay;
            s/^\[(.*)\]$/$1/ for (@relay);

            if($self->{_pmxpe_internal_hosts}->match(@relay)) {
                $msg->set_attr('mark_i', undef);
                return 'Outbound';
            }
        }
    }

    return 'Inbound';
}

sub get_paths {
    my $self = shift;
    my $msg = shift;

    $msg->template_object($self->{pmxpe_template});

    # refresh the members-per-group cdb if necessary
    $self->_refresh_group_map
        if time - $self->{pmxpe_last_refresh} > $self->{pmxpe_list_refresh_interval};

    # get the recipients
    my $minfo = $msg->minfo;
    my $sender = $minfo->value('Envelope-From');
    my @recipients = $minfo->value('Envelope-To');

    # Sort the recipients into buckets each bucket is a unique path
    # and requires a separate message object
    my %out_paths;
    my @instrs = @{$self->{pmxpe_per_recipient_instrs}};

    # get the user preferences if we don't have them already
    my $pref = $self->{user_preferences} ||= PureMessage::Preferences->new;
    my $splitting = 0;

    eval {
        # Initialize minfo field 'Direction'
        my $direction = $self->get_direction($msg);
        $minfo->value('Direction', $direction);
        $msg->log('DEBUG', "POLICY ENGINE: direction is $direction");

        my $outbound = ($direction eq 'Outbound') ? 1 : 0;
        my @groups = ();
        if ($self->{_pmxpe_gmap} && $outbound) {
            my $group = $self->{_pmxpe_gmap}->lookup($sender);
            if($group){
                $minfo->append_items('Sender-Group', "$sender|$group");
                push @groups, "$sender|$group";
            }
        }

        my $group_splitting = 0;
        foreach my $recip (@recipients) {
            my $mrecip   = $self->map_recipients($msg, $recip);
            my $language = $pref->language($mrecip);
            my $group;

            # set to a single recipient
            $minfo->value('Envelope-To', $recip);
            $minfo->value('Mapped-Recipients', $mrecip);

            if($self->{_pmxpe_gmap} && !$outbound) {
                $group = $self->{_pmxpe_gmap}->lookup($mrecip);
                if($group) {
                    $minfo->append_items('Recipient-Group', "$mrecip|$group");
                    push @groups, "$mrecip|$group";
                    $group_splitting = 1;
                }
            }

            # If for multiple recipients OR for a single recipient associated 
            # to a group
            if (@recipients > 1 || $group_splitting) {
                $splitting = 1;
            }

            if ($splitting) {
                $msg->log('DEBUG', "POLICY ENGINE: begin recipient-specific tests for '$mrecip'") 
                    if @instrs;

                my $path;
                foreach my $in (@instrs) {
                    my $instr = $self->{pmxpe_instructions}[$in];
                    my $rc = undef;
                    
                    # Execute instructions for tests
                    $rc = $self->_execute($instr, $msg);
                    $path .= $rc ? 1 : 0;
                }

                # default path
                $path = '0' unless defined $path;

                # A group name of '0' is not valid. We use it here to indicate the
                # lack of a group.
                $group = '0' unless $group;

                # put the recipient in a bucket
                push @{$out_paths{"$path;group:$group;lang:$language"}[0]}, $recip;
                push @{$out_paths{"$path;group:$group;lang:$language"}[1]}, $mrecip;
            }
            else {
                $minfo->value('User-Language', $language) if $language;
            }
        }
        $msg->set_attr('mark_g', \@groups) if scalar @groups;

        # restore original recipients
        $minfo->value('Envelope-To', @recipients);
    };

    if ($@) {
        # Trap Message Exception errors so that we can quarantine the message.
        if ($@ =~ /^MESSAGE EXCEPTION/) {
            my ($exc, $partid) = @{$msg->{base}}{qw(exception exception_part)};
            $msg->log('ERR', "POLICY STOP: message exception $exc at part $partid when splitting");
            $msg->delivery_action($self->{pmxpe_message_exception_action}, 
                                  'MessageException');
            $msg->stop;
            $self->_reset;
            return;
        }
        else {
            die $@;
        }
    }

    $msg->log('DEBUG', "POLICY ENGINE: end of recipient-specific tests")
        if (@instrs && $splitting);

    return %out_paths;
}

sub one_recipient_per_message {
    my $self = shift;
    return @{$self->{pmxpe_per_recipient_instrs}} > 0;
}

sub loaded_plugins {
    my $self = shift;
    return @{$self->{pmxpe_loaded_plugins}};
}

sub process {
    my $self = shift;
    my $msg = shift;
    my %opts = @_;

    # if called with a "path" argument, it means we are processing
    # one copy of a split message.
    my (@path,$path);
    if (exists $opts{path}) {
        $path = $opts{path};

        # Strip off the group & language information so that it can be
        # added to the message's minfo and not interfere with the normal
        # path processing.
	$path =~ s/;group:.*;lang:(.*)$//;
	my $language = $1;
	$msg->minfo_value("User-Language", $language) if $msg;
        @path = split '', $path;
    }

    _croak("POLICY ERROR: init() must be called before process()")
        unless $self->{pmxpe_init};

    _croak("POLICY ERROR: message object undefined!")
        unless $msg;

    # return if the message is done processing
    return if $msg->stopped;

    # reset the engine state if this is a new message
    if ($opts{new} or $self->{pmxpe_pc} == -1) {
        $self->_reset;
        $msg->context_reset;
        $self->{pmxpe_pc} = 0;
    }
    else {
        # else we are re-entering after an assert_event().  make sure the
        # command we are re-entering knows that.
        $self->{pmxpe_first_time} = 0;
    }

    # check if the event we were waiting for has happened yet
    delete $self->{pmxpe_assert_event} if $self->{pmxpe_assert_event}
        and event_passed($msg->event, $self->{pmxpe_assert_event});

    # if it hasn't, return
    return if $self->{pmxpe_assert_event};

    # refresh the lists if necessary
    $self->_refresh_lists
        if time - $self->{pmxpe_last_refresh} > $self->{pmxpe_list_refresh_interval};

    # give the message our template expander
    $msg->template_object($self->{pmxpe_template});

    # make the message an argument to template var functions when they are called
    $self->{pmxpe_template}->template_args([$msg]);

    # the index of the last instruction in the script (the final stop command)
    my $last_instr = $#{$self->{pmxpe_instructions}};

    my $stop = 0;
    eval {
        do {
            # lookup the instruction
            my $instr = $self->{pmxpe_instructions}[$self->{pmxpe_pc}];

            # log a rule hit
            my $rule = $instr->{rule};
            $msg->log('DEBUG', "POLICY RULE HIT: $rule")
                if $rule and $self->{pmxpe_first_time};

            # if we hit a stop command
            if ($instr->{name} eq 'stop') {

                # stop processing
                $msg->stop;

                # safety measure - make sure we stop even if the message object
                # does not report itself as being stopped, so that we don't loop
                # forever.
                $stop = 1;
            }
            # if the instruction is a test
            elsif ($instr->{type} eq 'test') {
                my $rc;

                # update the message context with if we are entering a new block
                # this happens when we hit an "if" or "elsif".
                $msg->context_push if $instr->{cpush};

                # update the message context with the state of negation
                # of the test. test = 0, not test = 1,  not not test = 0, etc
                $msg->context_not($instr->{cnot});

                # if we are in a message splitting situation, we want the
                # path variable to determine the return codes of any
                # per-recipient tests in the script.  we don't execute
                # these per-recipient tests more than once.
                my $found = 0;
                if (@path and $instr->{per_recipient}) {
                    for(my $i = 0; $i < @{$self->{pmxpe_per_recipient_instrs}}; $i++) {
                        if ($self->{pmxpe_pc} == $self->{pmxpe_per_recipient_instrs}[$i]) {
                            warn "POLICY WARNING: invalid path: $path\n", last unless defined $path[$i];
                            $rc = $path[$i];
                            $found++;
                            last;
                        }
                    }
                    warn "POLICY WARNING: invalid path: $path\n" unless $found;
                }

                # execute the test (unless its result was specified in the @path)
                if ($found) {
                    $msg->log('DEBUG', $instr->{logstr} . " returned "
                            . ($rc ? "true" : "false") . " (used cached result)");
                }
                else {
                    $rc = $self->_execute($instr, $msg, $self->{pmxpe_first_time});
                }

                # pop the context stack
                my $cpop = $rc ? $instr->{cpopt} : $instr->{cpopf};
                if ($cpop) {
                    $msg->context_pop for (1 .. $cpop);
                }

                # move the PC to the next instruction
                $self->{pmxpe_pc} = $rc ? $instr->{next_true} : $instr->{next_false};
            }

            # else the instruction must be an action
            else {

                $self->_execute($instr, $msg, $self->{pmxpe_first_time})
                    unless $instr->{name} eq '_placeholder_';

                # pop the context stack (cpopt = cpopf for actions)
                if ($instr->{cpopt}) {
                    $msg->context_pop for (1 .. $instr->{cpopt});
                }

                # move the PC to the next instruction (next_true and next_false
                # are always the same for actions)
                $self->{pmxpe_pc} = $instr->{next_true};
            }

            # first time for any subsequent commands
	    $self->{pmxpe_first_time} = 1;

        } until $msg->stopped or $stop;

    };

    if ($@) {

        # look for "ENGINE EVENT" - this means that a command needs more message
        # data to continue
	if ($@ =~ /^ENGINE EVENT/) {
	    my $event = (split(' ', $@))[2];
	    $self->{pmxpe_assert_event} = $event;
	}
	# look for "MESSAGE EXCEPTION" - this means that a message violated a
	# constraint.
	elsif ($@ =~ /^MESSAGE EXCEPTION/) {
	    my ($exc, $partid) = @{$msg->{base}}{qw(exception exception_part)};
	    $self->log('ERR', "POLICY STOP: message exception $exc at part $partid");
        $msg->delivery_action($self->{pmxpe_message_exception_action}, 
                              'MessageException');
	    $msg->stop;

	    # This code is copied from below so that I don't need a goto()
	    # into the cleanup code.
	    $self->_reset;
	    return 1;
	}
	else {
            my $instr = $self->{pmxpe_instructions}[$self->{pmxpe_pc}];
            $self->_reset;
            _croak("$instr->{name}(): ", $@);
        }
    }
    else {

        # log the fact that we stopped
        if ($self->{pmxpe_pc} == $#{$self->{pmxpe_instructions}}) {
            $msg->log('DEBUG','POLICY STOP: message finished processing (end of policy script)');

            # sanity
            _croak("Error: unstoppable message object!") unless $msg->stopped;
        }
        else {
            $msg->log('DEBUG','POLICY STOP: message finished processing');
        }

        $self->_reset;

        return 1;
    }
}

sub _reset {
    my $self = shift;
    # reset some variables
    $self->{pmxpe_template}->template_reset();
    delete $self->{pmxpe_assert_event};
    $self->{pmxpe_pc} = -1;
    $self->{pmxpe_first_time} = 1;
}

sub _execute {
    my $self = shift;
    my $instr = shift;
    my $msg = shift;
    my $first_time = shift;
    my $recip = shift;

    # set the first_time variable in the args
    # this is a deprecated interface that should be removed eventually
    # the new way is to get it as the third parameter (see below)

    if (defined $first_time) {
        # only call this if we have to
        $instr->{args}->first_time($first_time);
    }
    else {
        $first_time = 1;
    }

    $msg->log('DEBUG', $instr->{logstr}) if $first_time;

    # run the command, return the result
    my $rc = $self->{pmxpe_commands}{$instr->{name}}->(
                                                    $msg,
                                                    $instr->{args},
                                                    $first_time
                                                    );
    $msg->log('DEBUG', "POLICY TEST: $instr->{name} returned "
                . ($rc ? "true" : "false")
                . (defined $recip ? " for recipient '$recip'" : ""))
        if $instr->{type} eq 'test';

    return $rc;
}

sub _build_group_map {
    my ($self) = @_;
    eval {
        $self->{_pmxpe_gmap} = PureMessage::Group::CDBMap->Tie;
    };
    if($@) {
        $self->log('DEBUG', "Policy WARNING: Error loading member-per-group.cdb: '$@'. NO GROUP MAPPING WILL TAKE PLACE");
    }
}

sub _refresh_group_map {
    my ($self) = @_;
    $self->_build_group_map unless $self->{_pmxpe_gmap};
    $self->{_pmxpe_gmap}->refresh if $self->{_pmxpe_gmap};
}

sub _refresh_lists {
    my $self = shift;

    $self->{pmxpe_script}->refresh_lists;

    $self->_refresh_recipient_map;

    $self->_refresh_trusted_relays;

    $self->_refresh_internal_hosts;

    $self->{pmxpe_last_refresh} = time;

    $self->{user_preferences}->reload if $self->{user_preferences};
}

1;
