package PureMessage::Policy::Engine::FileDriver;
use strict;
use Carp qw(croak);
use PureMessage::Policy::Events;
use PureMessage::Policy::MessageObject;
use Net::Domain qw(hostfqdn);
use PureMessage::Policy::Engine::ResultSet;
use PureMessage::Mbox::Parser;
use PureMessage::Config '%config';

sub new {
    my $class = shift;
    my $engine = shift;
    my %opts = @_;
    $opts{engine} = $engine;
    $opts{fqdn}  = hostfqdn() || 'localhost';
    # make sure env_to is an array
    my @to = ref($opts{env_to}) ? @{$opts{env_to}} : $opts{env_to};
    $opts{env_to} = \@to;
    return bless(\%opts, $class);
}

sub inject_file {
    my $self = shift;
    my @files = @_;
    
    delete $self->{resultset};
    $self->_set_minfo_defaults;

    foreach my $file (@files) {
        die "Error: '$file' is a directory\n" if -d $file;
        die "Error: '$file' not found\n" unless -f $file;
    }
    
    foreach my $f (@files) {
        my $msg;
        my $mbox = 0;
        $self->{engine}->log('DEBUG',"ENGINE DRIVER: opening '$f'...");
        $self->{cur_file} = $f;
        $self->{cur_id} = 0;
        open INFILE, $f or $self->_croak("$f: $!");
        
        # decide whether or not the file is in mbox format
        LINE: while (<INFILE>) {
            if (/^$/) {
                next LINE;
            }
            elsif (/^From /) {
                $mbox = 1;
            }
            last LINE;
        }
        seek INFILE, 0, 0;
        my $bs = $config{block_size} || 8192;
        if ($mbox) {
            # init the mbox parser
            my $msg;
            my $parser = $self->{parser} ||= do {
                my $p = new PureMessage::Mbox::Parser;
                $p->handler('message_start' => sub {
                        $self->{cur_id}++;
                        $msg = $self->_new_message();
                    });
                $p->handler('message_data' => sub {
                        $msg->print($_[0]);
                    });
                $p->handler('message_end' => sub {
                        $self->_process_eom($msg);
                    });
                $p;
            };
            while (read(INFILE, my $buf, $bs)) {
                $parser->parse($buf);
            }
            $parser->eof;
        }
        else {
            $msg = $self->_new_message();
            while (read(INFILE, my $buf, $bs)) {
                $msg->print($buf);
            }
            $self->_process_eom($msg);
        }
        close INFILE;
    }
    return $self->{resultset};
}

sub inject_quarantine {
    my $self = shift;
    my @msgids = @_;

    delete $self->{resultset};
    my %override = ();
    
    foreach (qw(relay env_from fur)) {
        $override{$_} = 1 if $self->{$_};
    }
    $override{env_to} = 1 if defined $self->{env_to}->[0] and length $self->{env_to}->[0];
    
    $self->{cur_file} = 'Quarantine';
    $self->{cur_id} = 0;

    foreach my $id (@msgids) {
        $self->{cur_id} = $id;
        my $qmsg = $self->{quarantine}->message($id)
            or $self->{engine}->log('WARNING',"Message $id not found, skipping.\n"), next;
        $self->{engine}->log('WARNING', "Message $id is in the trash, skipping.\n"), next
            if $qmsg->folder eq 'trash';
        next if $qmsg->minfo_value('reason') eq 'Queue'
            or $qmsg->minfo_value('reason') eq 'Test';
        $self->{relay} = $qmsg->minfo_value('relay') unless $override{relay};
        $self->{fur} = $qmsg->minfo_value('fur') unless $override{fur};
        $self->{env_from} = $qmsg->minfo_value('envelope-from') unless $override{env_from};
        $self->{env_to} = [$qmsg->minfo_value('envelope-to')] unless $override{env_to};
        
        my $msg = $self->_new_message();
        
        my $buf;
        while ($qmsg->read($buf, 2048)) {
            $msg->print($buf);
        }
        $self->_process_eom($msg);
    }
    return $self->{resultset};
}

sub inject_data {
    my $self = shift;
    my $data = shift;
    
    delete $self->{resultset};
    $self->{cur_file} = 'Data';
    $self->{cur_id} = 0;
    $self->_set_minfo_defaults;
    my $msg = $self->_new_message();
    $msg->print($data);
    $self->_process_eom($msg);
    return $self->{resultset};
}

sub _set_minfo_defaults {
    my $self = shift;
    $self->{relay} ||= "$self->{fqdn} [127.0.0.1]";
    $self->{fur} ||= 'none';
    $self->{env_from} ||= 'pmx@'.$self->{fqdn};
    $self->{env_to}->[0] = 'puremessage-test@'.$self->{fqdn}
        unless defined $self->{env_to}->[0] and length $self->{env_to}->[0];
}

sub _process_eom {
    my $self = shift;
    my $main_msg = shift;
    my $engine = $self->{engine};
    
    $main_msg->event('eom');
   
    # get all possible paths through the per-recipient tests of the script
    # for this message
    my %paths = $engine->get_paths($main_msg);
    my @path_keys = keys %paths;
    
    if (@path_keys) {
        # create additional message objects
        for (my $i = 0; $i < @path_keys; $i++) {
            
            my $path = $path_keys[$i];
            my $recips = $paths{$path};
            
            # split off one message per path
            my $msg = $main_msg->clone('Envelope-To', $recips->[0], 'Mapped-Recipients', $recips->[1]);
                        
            # process the message
            $self->_process_msg($msg, $path);
        }
    }
    else {
        $self->_process_msg($main_msg);
    }
}

sub _process_msg {
    my $self = shift;
    my $msg = shift;
    my $path = shift;
    my $engine = $self->{engine};
    my %opts;
    
    my $recip = join ', ', $msg->envelope_recipients;
    
    $engine->log('DEBUG', "POLICY MILTER: processing message for '$recip'");

    $opts{'new'}  = 1;
    $opts{'path'} = $path if defined($path);

    $engine->process($msg, %opts);

    $engine->log('ERR', "POLICY MILTER: error: expected the engine to have stopped") unless $msg->stopped;
    
    $msg->delivery_action or $msg->delivery_action('keep');
    my $action = $msg->delivery_action()->{action};
    my $reason = $msg->delivery_action()->{reason};
    
    $reason = 'Queue' if $action eq 'keep';
    
    $self->{resultset}->file($msg,
                            action=>$action,
                            reason=>$reason,
                            );        
    $engine->log('DEBUG', "ENGINE DRIVER: message for '$recip' finished with action '$action'".($reason ? " (reason: $reason)" : ""));
    if ($action eq 'quarantine') {
        $engine->log('DEBUG', "ENGINE DRIVER: message quarantined for '$recip' with reason '$reason'");
    }
    elsif ($action eq 'keep'){
        $engine->log('DEBUG', "ENGINE DRIVER: message for '$recip' queued for delivery");
    }
    else {
        # do nothing...
    }
    $msg->free;
}

sub _new_message {
    my $self = shift;
    my $msg = PureMessage::Policy::MessageObject->new(
            quarantine => $self->{quarantine} || "",
            debug => $self->{debug},
            logger => $self->{engine}->logger,
    );
    $msg->minfo_value('Envelope-From',$self->{env_from});
    my @to = ref($self->{env_to}) ? @{$self->{env_to}} : $self->{env_to};    
    $msg->minfo_value('Envelope-To', @to);    
    $msg->minfo_value('Relay',$self->{relay});
    $msg->minfo_value('Fur',$self->{fur});
    my @time = split(' ', gmtime);
    $msg->minfo_value('Date', sprintf("%s, %02d %s %4d %s GMT", @time[0,2,1,4,3]));
    $msg->minfo_value('Sendmail-Host', 'localhost');
    $msg->event('envelope');
    
    if (!$self->{resultset}) {
        $self->{resultset} =
            PureMessage::Policy::Engine::ResultSet->new(
                quarantine => $self->{quarantine},
                delete => $self->{delete},
            );
    }
    
    $self->{resultset}->make_result_mi(
                    $msg,
                    src_file=>$self->{cur_file},
                    src_id=>$self->{cur_id},
                    dry_run => $self->{dry_run},
    );
    
    return $msg;
}

sub _croak {
    $| = 1; $| = 0; # flush stdout first
    croak @_;
}

1;
