package PureMessage::Policy::Engine::ResultSet;
use strict;
use Carp qw(croak);
use PureMessage::Quarantine;

# use message_log encoding to serialize test data
use PureMessage::MessageLog qw(split_line join_line);

BEGIN {
   package PureMessage::MessageStore::Dir;
   sub TEST_MODE() { 1; }
}

my $count = 0;

sub new {
    my $class = shift;
    my %opts = @_;
    $opts{quarantine} ||= PureMessage::Quarantine->new();
    $opts{id} ||= "test_result-". time . "-" . PureMessage::_sys::tid() . $count++;
    $opts{r} = [];
    return bless \%opts, $class;
}

sub file {
    my $self = shift;
    my $msg = shift;
    my %opts = @_;
    
    _store_testdata($msg, 'action', $opts{action});
    $msg->file($opts{reason});
}

sub make_result_mi {
    my $self = shift;
    my $msg = shift;
    my %opts = @_;
    if ($opts{dry_run}) {
        $opts{dry_run} = "1"
    }
    else {
        delete $opts{dry_run};
    }
    _store_testdata($msg, %opts);
    $msg->minfo_value('Queue-Id',$self->{id});
}

sub _store_testdata {
    my $msg = shift;
    my $test_data_line = $msg->minfo_value('PMX-Test');
    my (undef, @test_data) = split_line(defined $test_data_line ? $test_data_line : "");
    push @test_data, @_;
    $msg->minfo_value('PMX-Test', join_line(42, @test_data)); # arbitrary timestamp
}

sub _retrieve_testdata {
    my $test_data_line = shift;
    my (undef, @test_data) = split_line(defined $test_data_line ? $test_data_line : "");
    return @test_data;
}

sub _populate {
    my $self = shift;
    my $q = $self->{quarantine};

    # XXX there should be a way to recover if these index actions fail. For
    # example, they may timeout if the indexer is backlogged and doesn't give 
    # up the index lock in time.
    $q->index;
    $q->meta_index;
    my $mset = $q->message_set(where=>"m_queue_id = '".$self->{id}."'",
                                fields => [
                                            'm_reason',
                                            'm_envelope_to',
                                            'm_pmx_test',
                                            ],
    );
    my $i = 0;
    while (my $row = $mset->fetch($i++)) {
        my $id = ($row->message_ids)[0];
        my @fields = $row->fields;        
        my %test_data = _retrieve_testdata($fields[2]);
        push @{$self->{r}}, 
            PureMessage::Policy::Engine::Result->new(
                id => $id,
                reason => $fields[0],
                to => $fields[1],
                %test_data,
                quarantine => $self->{quarantine},
            );
        if ($test_data{dry_run}) {
            my $msg = $q->message($id);
        }
    }
    $self->{populated} = 1;
}

sub results {
    my $self = shift;
    $self->_populate() unless $self->{populated};
    return @{$self->{r}};
}

sub free {
    my $self = shift;
    while (my $r = shift @{$self->{r}}) {
        $r->free;
    }
}

sub id {
    return shift->{id};
}

1;

package PureMessage::Policy::Engine::Result;
use strict;
use Carp qw(croak);
use PureMessage::Quarantine;

sub new {
    my $class = shift;
    my %opts = @_;
    $opts{quarantine} ||= PureMessage::Quarantine->new();
    return bless \%opts, $class;
}

sub action {
    return shift->{action} || "copy to quarantine";
}

sub reason {
    return shift->{reason};
}

sub trace {
    croak "not implemented";
}

sub message {
    my $self = shift;
    return undef unless $self->{id};
    $self->_open_message unless $self->{message};
    return $self->{message};
}

sub id {
    return shift->{id};
}

sub source_file {
    return shift->{src_file} || '';
}

sub to {
    return shift->{to} || '';
}

sub source_id {
    return shift->{src_id} || 0;
}

sub log {
    croak "not implemented";
}

sub free {
    my $self = shift;
    #if ($self->{dry_run} or ($self->{action} ne 'quarantine'
    #                         and $self->{action} ne 'keep')) {
    #    $self->_open_message unless $self->{message};
    #    $self->{message}->trash unless $self->{message}->folder eq 'trash';
    #}
}

sub _open_message {
    my $self = shift;
    $self->{message} = $self->{quarantine}->message($self->{id});
}

1;


__END__

=head1 NAME

PureMessage::Policy::Engine::ResultSet - Get results from FileDriver injection

=head1 SYNOPSIS

my $rs = $file_driver->inject_file('foo');

my @results = $rs->get_results();

foreach my $result (@results) {
    print "Source file: ", $result->src_file, "\n";
    print "Message index: ", $result->src_id, "\n" if $result->src_id;
    print "Action: ", $result->action, "\n";
    print "Reason: ", $result->reason, "\n" if $result->action eq 'quarantine';
    print "Recipient: ", $result->to, "\n";
    my $msg = $result->message;
    # do stuff with MessageFile object (print source, etc.)
}

# You can save the id and then go back and get the result set again

my $rs_id = $rs->id;
my $rs = PureMessage::Policy::Engine::ResultSet->new($rs_id);

# When you are done, clean up.

$rs->free;
