package PureMessage::Policy::MessageText;

use strict;
use warnings;
use Sophos::SAVI::Engine qw($NOTIFY_DEFAULT $NOTIFY_STOP $NOTIFY_CONTINUE_NEXT);
use PureMessage::Policy::Phrase::Config qw(%phrase_config);
use PureMessage::Sophos::SAVI::AttachmentScan qw(get_scanner);
use PureMessage::Policy::Plugin::Utils qw(get_attachments
        get_attachment_filter_types
        get_attachment_filter_extensions
        get_attachment_filter_filetypes
        );
use PureMessage::Sophos::SAVI::TFT qw(get_tft_mime_types_mapping);
use File::Basename;
use PureMessage::Sophos::SAVI::Utils qw(set_savi_cantscan_errors);

use ActiveState::I18N::UTF8 qw(utf8_encode);
use PureMessage::UTF8::Config qw(%utf8_config);

use base qw(Exporter);
our @EXPORT_OK = qw(match_attachment_text 
        per_attachment_byte_limit_exceeded 
        per_message_time_limit_exceeded
        current_attachment_part_id
        );
    
my $current_part_id;

sub current_attachment_part_id { return $current_part_id }

sub match_attachment_text {
    my $msg = shift;
    my $matcher = shift;
    my %opts = @_;    
    my $matching_parts     = delete $opts{matching_parts};
    my $non_matching_parts = delete $opts{non_matching_parts};
    my $scan_all_attachments       = delete $opts{no_early_out};

    my $mime_types_inside;

    my $byte_count = 0;
    my @matches;
    my $current_part_ugly_name;
    my $current_part_pretty_name;
    my $current_file;
    my $current_tft_filename;
    my $limit_exceeded = 0;
    my $early_out      = 0;
    my $message_start_scan_time = time();

    my $map = $msg->get_attr('cantscan_map');
    my $cantscan_map = $map->[0];

    my $debug = 0;
    if (my $dir = $ENV{PMX_ATTACHMENT_SCAN_DEBUG_PATH}) {
        $debug = $dir if -d $dir;   
    }
    my $d_count = 1;

    # The keys of attachment_filters *must* match the keys returned by get_tft_mime_types_mapping() 
    my %attachment_filters;
    if (my @mime_type_filter = get_attachment_filter_types($msg)) {
        $attachment_filters{mime_types} = \@mime_type_filter;
        $msg->log('DEBUG', "messagetext: filtering by mime types: " . join(', ', @mime_type_filter));
    }
    if (my @extension_filter = get_attachment_filter_extensions($msg)) {
        $attachment_filters{extensions} = \@extension_filter;
        $msg->log('DEBUG', "messagetext: filtering by extensions: " . join(', ', @extension_filter));
    }
    if (my @filetype_filter = get_attachment_filter_filetypes($msg)) {
        $attachment_filters{filetypes} = \@filetype_filter;
        $msg->log('DEBUG', "messagetext: filtering by filetypes: " . join(', ', @filetype_filter));
    }

    my $part_type_mapping = { };

    if ($debug) {
        my $FH;
        my $debug_file = "$debug/0-debug-info.txt";
        my $buffer_size = $ENV{PMX_AV_CONTENT_BUFSIZE} || 4096;
        open $FH, ">$debug_file" or warn "Error opening '$debug_file': $!";
        print $FH "Buffer size: $buffer_size\nByte limit: $phrase_config{per_attachment_byte_limit}\n";
        close $FH or warn "Error closing '$debug_file': $!";
    }

    my $scanner = get_scanner({
	OnFileFound => sub {
        # Check if the time limit has been exceeded
        return $NOTIFY_STOP if ($limit_exceeded);

	    $current_file = shift;
        $current_tft_filename = basename($current_file);
	    $current_file =~ s#^\Q$current_part_ugly_name\E#$current_part_pretty_name#;
 
	    return $NOTIFY_DEFAULT;
	},

	OnStreamFound => sub {
        my $stream          = $_[0];
        my $stream_type     = $_[1];
        $limit_exceeded     = $_[2]; 

        my $stream_encoding = $_[3];
        $stream_encoding =~ s/(?:\W*)$//;
        $stream_encoding = 'UTF-8' if $stream_encoding eq 'Unknown';

        # set when the engine reaches its byte limit
        if (defined $limit_exceeded) {
            # Set the 'cantscan' reason using the SAVI ID and the part id
            if($limit_exceeded and (defined $cantscan_map)) {
                $msg->log('DEBUG', "messagetext: byte limit exceeded on part '$current_part_id'");
                push(@{$cantscan_map->{ids}},
                     {'sav_id' => 'SOPHOS_SAVI_MAX_BYTE_LIMIT_EXCEEDED', 'part_id' => "$current_part_id"} );
                return $NOTIFY_STOP;
            }
        }

        # Check the current running time
        if (!$early_out and ((time() - $message_start_scan_time) >= $phrase_config{per_message_time_limit})) {
            # Set the 'cantscan' reason using the SAVI ID and the part id
            if(defined $cantscan_map) {
                push(@{$cantscan_map->{ids}},
                     {'sav_id' => 'SOPHOS_SAVI_FILE_TIMEOUT', 'part_id' => "$current_part_id"} );
            }
            # Time Limit exceeded 
            # Use 'T' to differentiate this from a byte limit timeout

            $msg->log('DEBUG', "messagetext: time limit ($phrase_config{per_message_time_limit}) exceeded on part '$current_part_id'");
            $limit_exceeded = 'T';
            $early_out = 1;
            # The engine does not actually pay attention to the 'NOTIFY_STOP' return status
            # so we set the flags above and return a valid stop signal in 'OnFileFound'
            return $NOTIFY_STOP;
        }

        # See if the current filetype is in the list of types to be scanned

        foreach my $f (keys %attachment_filters) {
            #XXX: this should be a list matcher
            my @filter_types = @{$attachment_filters{$f}};
            $msg->log('DEBUG', "messagetext: mapping $f for file '$current_tft_filename': "
                      . join(',', @{$part_type_mapping->{$current_tft_filename}->{$f}}));
            return $NOTIFY_DEFAULT if(!(grep { my $x=$_;
                                               grep $x eq $_, @filter_types;
                                             }
                                             @{$part_type_mapping->{$current_tft_filename}->{$f}}));
        }

        # Transcode the stream to UTF-8 
        utf8_encode($stream, $stream_encoding, \%utf8_config)
            unless $stream_encoding eq 'UTF-8';

            my @m = $matcher->($stream);
        if ($m[0]) {
            push @matches, {
                file => $current_file,
                type => $stream_type,
                text => $m[1],
            } if $m[0] != -1;
            
            # early_out flags that we had a match
            $early_out = 1;
        }

        if ($debug) {
            my $FH;
            my $sanitized_file = $current_file;
            $sanitized_file =~ s#/#_#g;
            my $debug_file = "$debug/$d_count-$sanitized_file";
            open $FH, ">$debug_file" or warn "Error opening '$debug_file': $!";
            print $FH $stream;
            close $FH or warn "Error closing '$debug_file': $!";
            $d_count++;
        }

        return $early_out ? $NOTIFY_STOP : $NOTIFY_DEFAULT;
       }},
    content_scanning_per_attachment_byte_limit => $phrase_config{per_attachment_byte_limit},
    );


    my @filtered = get_attachments($msg);
    my @parts = $msg;
    my $part;
    while ($part = shift @parts) {
        my @ids = $part->parts;
        unshift @parts, map { $part->part($_) } @ids if @ids;
        
        next if @filtered && !(grep $part->part_id eq $_, @filtered);

        $current_part_id = $part->part_id;

        if ($non_matching_parts) {
            # Initially assume this part is non-matching
            push @$non_matching_parts, $current_part_id;
        }

        # If we've already had a match and we don't want to scan all attachments,
        # stop here: all we're doing is pushing the rest of the parts onto
        # the non-matching parts array
        if ($early_out && !$scan_all_attachments) {
            $msg->log('DEBUG', "messagetext: not scanning part '$current_part_id'");
            next;
        }

        my $pretty = PureMessage::Policy::Bundle::_get_names($part);

        $part_type_mapping = get_tft_mime_types_mapping($part);

        $current_part_ugly_name = $part->body_filename;
        $current_part_pretty_name = @$pretty ? $pretty->[0] : $current_part_ugly_name;
        $msg->log('DEBUG', "messagetext: scanning attachment '$current_part_pretty_name'");
       
        # Scan file and check for any SAVI errors
        $early_out  = 0; # reset early_out before each scan 
        my $errors  = $scanner->scan_file($current_part_ugly_name);
        my $sav_ids = $errors->{$current_part_ugly_name};

        $msg->{per_message_savi_cantscan_errors} += set_savi_cantscan_errors($msg, $current_part_id, $sav_ids);
    
        if ($early_out || @$sav_ids) {
            # If early_out is set or there were errors, 
            # pmx_phrase returns a positive result.
            # Therefore, save this part id as matching and remove it 
            # from the collection of non-matching part ids.      
            push @$matching_parts, $current_part_id if $matching_parts;
            pop @$non_matching_parts if $non_matching_parts;
        }    
        
        # We may have modified cantscan_map in the above set_savi_errors() call  
        $map = $msg->get_attr('cantscan_map'); 
        $cantscan_map = $map->[0];

        # We can only jump out if we've found something and we're not trying to
        # scan all attachments
        last if $early_out && !$scan_all_attachments;
    }

    if ($limit_exceeded and !@matches) {
        if($limit_exceeded eq 'T') {
            $msg->{per_message_time_limit_exceeded} = 1;
            $msg->log('DEBUG', "messagetext: message time limit ($phrase_config{per_message_time_limit}) exceeded");
        }
        else {
            $msg->{per_attachment_byte_limit_exceeded} = 1;
            $msg->log('DEBUG', "messagetext: message per attachment byte limit ($phrase_config{per_attachment_byte_limit}) exceeded");
        }
        return ();
    }

    return @matches;
}

sub per_attachment_byte_limit_exceeded {
    return exists shift->{per_attachment_byte_limit_exceeded} ? 1: 0;
}

sub per_message_time_limit_exceeded {
    return exists shift->{per_message_time_limit_exceeded} ? 1: 0;
}

1;
