package PureMessage::Policy::Repository::File;
use strict;
use PureMessage::Sieve qw(sieve_parse_file sieve_print sieve_text);
use PureMessage::Version qw($PREFIX);
use PureMessage::Policy::Repository::Config qw(%policy_repository_config);
use File::Basename qw(basename);
use Data::Dumper;
use ActiveState::File::Atomic;

my $DEFAULT_EMPTY_FILE = <<EOF;
# attr NAME=new file
stop;
EOF

=pod

use Data::Dumper;
use strict;

my $data = PureMessage::Policy::File->Available();
$data->[0]->description('how now brown cow');

$data = PureMessage::Policy::File->new('p1.siv');
print Dumper($data);
print "got: " . $data->description();

=cut

=item Load

Attempt to load an existing file from the respository.

=cut

sub Load {
  my $CLASS = shift;
  $CLASS = ref($CLASS) || $CLASS;
  my $self;
  $self->{file} = shift;
  my %opts = @_;
  $self->{dir} = $opts{dir} || $policy_repository_config{dir};
  bless($self,$CLASS);
  return unless $self->_info();
  return $self;
}

=item Create($file[,$contents])

Attempt to create a new file called $file. If $contents are passed
then it is created with the passed string as it's contents.

The newly created file will have the following contents:

=over 5

# attr NAME=new file
stop;

=back 5

=cut

sub Create {
  my $CLASS = shift;
  $CLASS = ref($CLASS) || $CLASS;
  my $self;
  $self->{file} = shift || _available_name();
  $self->{dir} = $policy_repository_config{dir};
  bless($self,$CLASS);
  die "file exists" if -e $self->absfile();
  if (my $contents = shift) {
    _set_file_contents($self->absfile(),$contents);
  } else {
    _set_file_contents($self->absfile(),$DEFAULT_EMPTY_FILE); # make a new emptyish file
  }
  $self->_info();
  return $self;
}

=item set_file_contents($contents)

Replace file contents with $contents.

=cut

sub set_file_contents {
  my ($self,$contents) = @_;
  _set_file_contents($self->absfile(),$contents);
}

# this sub may die if there is an error
sub _set_file_contents {
  my ($file,$new_contents) = @_;
  my $fa = new ActiveState::File::Atomic($file,writable => 1,create => 1);
  my $temp_fh = $fa->tempfile();
  print $temp_fh $new_contents;
  $fa->commit_tempfile();
}


#examine_sieve_file($filename)

#Returns a hash reference of information about the passed filename.

#Interesting keys are: desc, size, mtime


sub _examine_sieve_file {
  my ($filename) = @_;
  my $f;
  my @comments;
  my @requires;
  my @stat = stat($filename);
  return unless @stat;
  eval {
    $f = sieve_parse_file($filename);
    @requires = $f->requires();
    foreach my $c ($f->commands()) {
     push(@comments,$c->comment());
    }
  };

  if ($@) {
    warn $@;
    return; # not sure what happened but we couldn't read the file
  }
  my %data;
  foreach my $c (@comments) {
    if (my $r = extract_attr($c)) {
      $data{desc} = $r;
    }
  }
  $data{mtime} = $stat[9];
  $data{size} = $stat[7];
  $data{requires} = \@requires;
  $data{script} = scalar(@requires);
  $data{snippet} = $data{script} ? 0 : 1;
  $data{desc} = '[UNKNOWN]' unless $data{desc};
  return \%data;    
}

# extract_attr($line)

# Attempts to parse a comment line as found in a sieve file for PureMessage
# attributes embedded in the comment.

#Returns the NAME if attribute if found.

sub extract_attr {
  my $full_line = shift;
  # sieve parser will return only one comment for multi line
  # comments, keeping the embedded newlines
  my @all_comments = split("\n",$full_line);
  return unless scalar(@all_comments);
  foreach my $line (@all_comments) { 
    if ($line =~ /^attr[ \t]*NAME[ \t]*=[ \t]*(.*)$/ ) {
      return $1; 
    } 
  }
  return;
}



=item FromClipboard($filename)

Create a new file copying the contents using the lines in the 
clipboard as it's initial content.

=cut

sub FromClipboard {
   my $CLASS = shift;
  $CLASS = ref($CLASS) || $CLASS;
  my $self = _init();
  $self->{file} = shift;
  bless($self,$CLASS);
  $self->copy();
  return unless $self->{info};
  return $self;
}

sub file_exists {
  my $filename = shift;
  $filename = $filename =~ m|^/| ? $filename : "$policy_repository_config{dir}/$filename";
  return 1 if -e $filename;
}

sub _available_name {
  my $template = 'clip%03s.siv';
  my $index = 0;
  my $fname = sprintf($template,$index);
  while(file_exists($fname)) {
    $index++;
    $fname = sprintf($template,$index);
  }
  return $fname;
}

sub contents {
  my $self = shift;
  my $content;
  {
    open FV,$self->absfile() || return;
    local $/ = undef;
    $content = <FV>;
    close FV;
  }
  return $content;
}

sub _info {
  my $self = shift;
  my $info = _examine_sieve_file($self->absfile);
  return unless $info;
  $self->{info} = $info;
  return $info;
}

sub info {
  my $self = shift;
  my $temp = $self->{info} ? $self->{info} : $self->_info();
  $temp = ref($temp) eq 'HASH' ? $temp : {};
  return wantarray ? %$temp : $temp->{desc};
}
sub file {
  my $self = shift;
  return basename($self->{file});
}

sub absfile {
  my $self = shift;
  return $self->{file} =~ m|^/| ? $self->{file} : "$self->{dir}/$self->{file}";
}

sub from_clipboard {
  my $self = shift;
  my $af = new ActiveState::File::Atomic($self->absfile(),writable => 1, create => 1);
  $af->commit_file("$PREFIX/etc/.clipboard.siv");
  $self->_info();
}

# this can die on error
sub from_file {
  my ($self,$file) = @_;
  if (ref($file) and $file->isa(ref($self))) {
    $file = $file->absfile();
  } elsif ($file !~ m|^/|) {
    $file = "$self->{dir}/$file";
  } 
  
  my $af = new ActiveState::File::Atomic($self->absfile(),writable => 1, create => 1);
  $af->commit_file($file);
  $self->_info();
}

sub rename {
  my $self = shift;
  my $new_location = shift;
  my $new_location_abs = $new_location =~ m(^/) ? $new_location : "$self->{dir}/$new_location";
  die "duplicate file" if -e $new_location_abs;
  rename $self->absfile,$new_location_abs;
  $self->{file} = $new_location;
}

sub to_clipboard {
  my $self = shift;
  my $af = new ActiveState::File::Atomic("$PREFIX/etc/.clipboard.siv",writable => 1, create => 1);
  $af->commit_file($self->absfile());
}

sub name {
  my $self = shift;
  my $name = basename($self->{file});
  $name =~ s/_/ /g;
  $name =~ s/.siv$//;
  return $name;
}

sub examine_sieve_file {
  my ($self, $file) = @_;
  $file = $self->absfile;
  $self->{file} = $file;
  return _examine_sieve_file($file);
}

sub delete {
  my ($self) = @_;
  unlink($self->absfile());
  $self = {};
}

sub description {
  my $self = shift;
  if (my $newdesc = shift) {
    if(my $old =  _update_sieve_file_description($self->absfile,$newdesc)) {
      $self->{info}->{desc} = $newdesc;
      return $old;
    }
    return;
  } else {
    return $self->{info}->{desc};
  }
}

# this sub may die if there is an error
sub _update_sieve_file {
  my ($file,$new_contents) = @_;
  my $fa = new ActiveState::File::Atomic($file,writable => 1);
  my $temp_fh = $fa->tempfile();
  print $temp_fh $new_contents;
  $fa->commit_tempfile();
}

# _update_seive_file_description($file,$desc)

#Attempts to update the embedded comment on the first sieve command 
#in this file. This is the comment that the examine_sieve_file function will
#find.


# this may croak on error
sub _update_sieve_file_description {
  my($file,$desc) = @_;
  
  my $f = sieve_parse_file($file);

  sub _set_comment {
    my ($file,$f,$desc,$first) = @_;
    my $found;
    foreach my $c ($f->commands()) {
      
      if(!$found) {
        if ( $c->can('comment') and ( $first or ($found = extract_attr($c->comment) and !$first))){
            
          $found = $c->comment unless $found; 
          if ($c->name eq 'if') {
            my @args = $c->args();
            $args[0]->comment("attr NAME=$desc");
          } else {
            $c->comment("attr NAME=$desc");
          }

          _update_sieve_file($file,sieve_text($f));
        }
      }
    }
    return;
  }
 
  my $old = _set_comment($file,$f,$desc) || _set_comment($file,$f,$desc,1);
  return $old if $old;
  return;
}


1;


