package PureMessage::Policy::Script;

use strict;
use Carp 'croak';
use PureMessage::Policy::Args;
use PureMessage::Policy::Template;
use PureMessage::ListManager;
use PureMessage::Policy::Plugin::Utils qw(find_multifiles);

sub new {
    my $class = shift;
    my $file = shift;
    my $plugins = shift;
    my %opts = @_;
    
    my $newcache = delete $opts{newcache};
    
    my $cache = "$file.cache";
    
    my $tree;
    
    if (-f $cache) {
        my $cache_time = (stat $cache)[9];
        my $script_time = (stat $file)[9];
        if ($script_time > $cache_time or $newcache) {
            $tree = _load_script($file, $plugins, %opts);
            _write_cache($tree, $cache);
        }
        else {
            eval{$tree = _read_cache($cache)};
            if ($@) {
                warn "Possibly corrupted $cache.  Fixing...\n";
                $tree = _load_script($file, $plugins, %opts);
                _write_cache($tree,$cache);
            }
        }
    }
    else {
        $tree = _load_script($file, $plugins, %opts);
        _write_cache($tree,$cache);
    }
    
    # build list objects for each instruction that uses them
    my %lists;
    my $multifiles = {};
    foreach my $instr (@$tree) {
        if ($instr->{args}) {
            $lists{$_} = 1 for $instr->{args}->find_named_lists;
            my ($id, $charset) = $instr->{args}->get_multifiles;
            $multifiles->{$id} = $charset if $id;
        }
    }

    # Check the existence of multifiles. The calls below will die if they
    # fail, which is what we want.
    find_multifiles($multifiles);
    
    my $lm = PureMessage::ListManager->new(preload=>[keys %lists]);

    foreach my $instr (@$tree) {
        if ($instr->{args}) {
            $instr->{args}->build_lists($lm);
            $instr->{per_recipient} = _is_per_recipient($lm, $instr);
        }
    }
    
    my $self = bless {tree=>$tree, lm=>$lm}, $class;
    
    return $self;
}

sub _is_per_recipient {
    my $lm = shift;
    my $in = shift;

    # envelope 'to' or 'group' test
    if($in->{name} eq 'envelope') {
        my $key = lc $in->{args}->param(0)->[0];
        return 1 if ($key eq 'to') or ($key eq 'group');
    }

    if (my @lists = $in->{args}->find_named_lists) {
        foreach my $list (@lists) {
            return 1 if $lm->list_exists($list) and $lm->is_multilist($list);
        }
    }
}

sub tree {
    return shift->{tree};
}

sub instruction_list {
    my $self = shift;
    return @{$self->{tree}};
}

sub instruction {
    my $self = shift;
    my $i = shift;
    return $self->{tree}->{$i};
}

sub command_hash {
    my $self = shift;
    
    my %command_hash = ();
    foreach my $instr (@{$self->tree}) {
        next if $instr->{name} =~ /^_/;
        $command_hash{$instr->{name}} = 0;
        next unless $instr->{templates};
        foreach my $template (@{$instr->{templates}}) {
            $command_hash{$template} = 0;
        }
    }
    return %command_hash;
}

sub template_hash {
    my $self = shift;

    my $template_hash = {};
    foreach my $instr (@{$self->tree}) {
        next unless $instr->{templates};
        foreach my $template (@{$instr->{templates}}) {
            $template_hash->{$template} = 0;
        }
    }
    return $template_hash;
}

sub required_modules {
    my $self = shift;
    
    my %modules;
    foreach my $instr (@{$self->tree}) {
        if ($instr->{module}) {
            $modules{$instr->{module}} = 0;
        }
    }
    return keys %modules;
}

sub refresh_lists {
    my $self = shift;
    
    $self->{lm}->refresh_lists;
    
    foreach my $instr (@{$self->tree}) {
        $instr->{args}->refresh_list if $instr->{args};
    }
    
}

sub _write_cache {
    my $o = shift;
    my $file = shift;
    
    require Data::Dump;
    
    open FILE, ">$file" or croak ("$file: $!");
    print FILE Data::Dump::dump($o);
    close FILE;
}

sub _read_cache {
    my $file = shift;
    
    my $o = do {require $file};
    
    return $o;
}

sub _load_script {
    my $file = shift;
    my $plugins = shift;
    
    require PureMessage::Policy::Script::Loader;
    my $loader = PureMessage::Policy::Script::Loader->new(plugins=>$plugins);
    
    my $script = $loader->load_script($file);

    return $script;
}

1;
