package PureMessage::Policy::Script::Loader;

use strict;
use Carp;
use PureMessage::Sieve qw(sieve_parse_file);
use PureMessage::Sieve::Nodes 'new_action';
use PureMessage::Policy::Config;
use ActiveState::UTF8;

sub _croak {
    $| = 1; $| = 0; # flush stdout first
    Carp::croak(@_);
}

sub new {
    my $self = shift;
    my $class = ref($self) || $self;
    my %opts = @_;
        
    _croak "missing 'plugins' parameter" unless exists $opts{plugins};
    
    my $plugins = $opts{plugins};
    
    my @plugins = ref $plugins ? @$plugins : ($plugins);
    
    my $conf = PureMessage::Policy::Config->new();
    
    # load plugins
    foreach my $plugin (@plugins) {
        $conf->load_plugin($plugin);
    }
    
    my %hash = (
        conf => $conf,
    );
    
    return bless \%hash, $class;
}

sub _init {
    my $self = shift;
    $self->{script} = [];
    $self->{paths} = 0;
    delete $self->{last};
}

sub load_script {
    my $self = shift;
    my $script_file = shift;
    my %opts = @_;
    
    $self->_init();
    
    # load sieve script
    my $tree = sieve_parse_file($script_file);
    ($self->{script_base} = $script_file) =~ s#^.*/##;

    # validate sieve "require" statement
    foreach my $req ($tree->requires) {
        _croak "POLICY ERROR: PureMessage does not provide '$req'\n"
            unless $self->{conf}->is_provided($req);
        $self->{requires}{$req} = 1;
    }
    
    $tree->commands($tree->commands, new_action('stop'));
    
    $self->_traverse_block($tree);
    
    do {
        delete @$self{'last', 'truth', 'tstack', 'rule'};
        @$self{'cpop', 'cpush', 'cnot'} = (0,0,0);
        $self->_link_block($tree);
    } while ($self->{paths} > 0);
    
    my $script = $self->{script};
    
    my $end_stop = $script->[$#$script];
    @$end_stop{'next_true', 'next_false', 'cpush', 'cpopf', 'cpopt', 'cnot'} = (0,0,0,0,0,0);

#   XXX Remove this optimization for now
#   XXX bring it back when it works fully (see the todos in loader.t)
#
#    $self->_prune_constants() unless $opts{no_prune};
    
    return $script;
}

sub _traverse_block {
    my $self = shift;
    my $block = shift;
    
    foreach my $command ($block->commands) {
        if ($command->name eq 'if') {
            foreach my $cond ($command->args) {
                if ($cond->test) {
                    $self->_traverse_tests($cond->test);
                }
                $cond->commands(new_action('_placeholder_')) unless $cond->commands;
                $self->_set_rule($cond);
                $self->_traverse_block($cond);
            }
        }
        else {
            $self->_add_instruction($command, 'action');
        }
    }
}

sub _traverse_tests {
    my $self = shift;
    my @tests = @_;
    
    foreach my $test (@tests) {
        if ($test->name eq 'not') {
            $self->_traverse_tests($test->args);
        }
        elsif ($test->name eq 'allof' or $test->name eq 'anyof') {
            $self->_traverse_tests(($test->args)[0]->tests);
        }
        else {
            $self->_add_instruction($test, 'test');
        }
    }
}

sub LTRUE() {1}
sub LFALSE() {2}

sub _link_block {
    my $self = shift;
    my $tree = shift;

    COMMAND: foreach my $command ($tree->commands) {
        if ($command->name eq 'if') {
            COND: for my $cond ($command->args) {
                if ($cond->test) {
                    my $commands = scalar $cond->commands;
                    $self->_context_push if $commands;
                    my $rc = $self->_link_test($cond->test);
                    $self->_link_block($cond) if $rc;
                    $self->_context_pop if $commands;
                    last COND if $rc;
                }
                else {
                    $self->_link_block($cond);
                    last COND;
                }
            }
            next COMMAND;
        }
        else {
            $self->_link_to_last($command, LTRUE | LFALSE);
            next COMMAND;
        }
    }  
}

sub _context_push {
    my $self = shift;
    _croak "cpush should be zero" if $self->{cpush}; # assert
    $self->{cpush} = 1;
}

sub _context_pop {
    my $self = shift;
    $self->{cpop}++;
    _croak "cpush should be zero" if $self->{cpush}; # assert
}

sub _context_flip {
    my $self = shift;
    $self->{cnot} = $self->{cnot} ? 0 : 1;
}

sub _context_pop_take {
    my $self = shift;
    my $retval = $self->{cpop};
    $self->{cpop} = 0;
    return $retval;
}

sub _context_push_take {
    my $self = shift;
    my $retval = $self->{cpush};
    $self->{cpush} = 0;
    return $retval;
}

sub _context_not {
    my $self = shift;
    return $self->{cnot};
}

sub _link_to_last {
    my ($self, $command, $newtruth) = @_;
    
    my $script = $self->{script};
    my $this = $command->line;
    my $this_in = $script->[$this];
    
    my $last = $self->{last};
    my $truth = $self->{truth};
    
    if (defined $last) {
        my $last_in = $script->[$last];
        my $cpop = $self->_context_pop_take;
        if ($truth & LTRUE) {
            $last_in->{cpopt} = $cpop;
        }
        if ($truth & LFALSE) {
            $last_in->{cpopf} = $cpop;
        }
        $self->_link($last, $this, $truth);
    }
        
    @$this_in{'cpush', 'cnot'} = (0,0) unless $this_in->{type} eq 'test';
 
    $self->{last} = $this;
    return $self->_truth($newtruth);
}

sub _truth {
    my $self = shift;
    my $newtruth = shift;
    $self->{truth} = $newtruth;
    return ($newtruth & LTRUE);
}

sub _link_test {
    my $self = shift;
    my $test = shift;
    
    my $test_name = $test->name;
        
    if ($test_name eq 'anyof') {
        foreach (($test->args)[0]->tests) {
            return 1 if ($self->_link_test($_));
        }
        return 0;
    }
    elsif ($test_name eq 'allof') {
        foreach (($test->args)[0]->tests) {
            return 0 unless ($self->_link_test($_));
        }
        return 1;
    }
    elsif ($test_name eq 'not') {
        $self->_context_flip;
        my $rc = not $self->_link_test($test->args);
        $self->_context_flip;
        return $rc;
    }
    
    my $script = $self->{script};
    my $this = $test->line;
    my $this_in = $script->[$this];
    my $rc;
    my $tstack = ($self->{tstack} ||= []);
    
    if (exists $this_in->{next_false}) {
        $rc = $self->_link_to_last($test, LFALSE);
    }
    elsif (exists $this_in->{next_true}) {
        if ($this_in->{paths} > 0) {
            $rc = $self->_link_to_last($test, LTRUE);
        }
        else {
            foreach (@$tstack) {
                my $in = $script->[$_];
                $in->{paths}--;
            }
            $self->{paths}--;
            $rc = $self->_link_to_last($test, LFALSE);
        }
    }
    else {
        foreach (@$tstack) {
            my $in = $script->[$_];
            $in->{paths}++;
        }
        $self->{paths}++;
        $this_in->{paths} = 0;
        $this_in->{cnot} = $self->_context_not();
        $this_in->{cpush} = $self->_context_push_take();
        $rc =  $self->_link_to_last($test, LTRUE);
    }
        
    $self->_context_push_take();
        
    push @$tstack, $this;
    
    return $rc;
}


sub _link {
    my ($self, $from, $to, $truth) = @_;
    
    my $script = $self->{script};
    
    return unless defined $from and $to > 0;
    
    if ($truth & LTRUE) {
        $script->[$from]{next_true} = $to unless exists $script->[$from]{next_true};
    }
    if ($truth & LFALSE) {
        $script->[$from]{next_false} = $to unless exists $script->[$from]{next_false};
    }
}

sub _add_instruction {
    my $self = shift;
    my $command = shift;
    my $type = shift;

    my %instruction;
    $instruction{type} = $type;
    $instruction{name} = $command->name;
    $instruction{visited} = 0;
    $instruction{line} = $command->line;
    $instruction{rule} = $self-> _get_rule();
    
    unless ($instruction{name} eq '_placeholder_') {
        $instruction{args} = $self->_validate($command);
        my $conf = $self->{conf};
        foreach my $t ($instruction{args}->templates) {
                push @{$instruction{templates}}, $t if $conf->is_of_type('template_var', $t);
        }
    
        $instruction{module} = $self->_what_provides($instruction{name});
    
        # create the string that gets printed to the log before
        # this command is executed
        my $logstr = $type eq 'test' ? "POLICY TEST: " : "POLICY ACTION: ";
        $logstr .= $command->name;
        foreach my $param ($instruction{args}->params) {
            my $pstr = "";
            if (defined $param) {
                if (ref $param) {
                    $pstr .= join ', ', @$param;
                }
                else {
                    $pstr .= $param;
                }
                if (length $pstr > 32) {
                    $pstr = substr $pstr, 0 , 32;
                    $pstr .= '...';
                }
                $pstr =~ s/[\r\n\t]+/ /g;
            }
            $logstr .= " \"$pstr\"" if $pstr;
        }
        $instruction{logstr} = $logstr;
    }
    
    my $script = $self->{script};
    push @$script, \%instruction;
    
    # store script index in tree node
    $command->line($#$script);
}

sub _prune_constants {
    my $self = shift;
    my $script = $self->{script};
    
    # trivial case - an empty script with only the stop command
    return unless @$script > 1;
    
    COMMAND: for (my $i = 0; $i < @$script; $i++) {
        my $in = $script->[$i];
        
        # if this is a 'true' test redirect any command that points
        # to it to its 'next_true' ptr
        if ($in->{name} eq 'true') {
            for (my $j = 0; $j < @$script; $j++) {
                next if $j == $i;
                $script->[$j]{next_true} =  $in->{next_true}
                    if $script->[$j]{next_true} == $i;
                $script->[$j]{next_false} =  $in->{next_true}
                    if $script->[$j]{next_false} == $i;
            }
            _make_into_placeholder($in);
            next COMMAND;
        }
        
        # if this is a 'false' test redirect any command that points
        # to it to its 'next_false' ptr
        if ($in->{name} eq 'false') {
            for (my $j = 0; $j < @$script; $j++) {
                next if $j == $i;
                $script->[$j]{next_true} =  $in->{next_false}
                    if $script->[$j]{next_true} == $i;
                $script->[$j]{next_false} =  $in->{next_false}
                    if $script->[$j]{next_false} == $i;
            }
            _make_into_placeholder($in);
            next COMMAND;
        }
        
        # prune any commands that have no-one pointing to them
        my $found;
        for (my $j = 0; $j < @$script; $j++) {
            next if $j == $i or $script->[$j]{name} eq '_placeholder_';
            $found = 1 if $script->[$j]{next_true} == $i
                or $script->[$j]{next_false} == $i
        }
        _make_into_placeholder($in) unless $found;
    }    
}

sub _make_into_placeholder {
    my $instr = shift;
    delete $instr->{$_} foreach (keys %$instr);
    $instr->{name} = '_placeholder_';
}

sub _validate {
    my $self = shift;
    my $node = shift;
    
    my $requires = $self->{requires};
    my $conf = $self->{conf};
    my $name = $node->name;
    unless (exists $requires->{'PureMessage'}) {
        foreach my $ns ($conf->namespaces) {
            if ($conf->is_in_namespace($ns,$name) and !$requires->{$ns}) {
                _croak "POLICY ERROR: You must require \"PureMessage\" to use \"$name\"\n" if $ns eq 'PureMessage';
                _croak "POLICY ERROR: You must require \"$ns\" or \"PureMessage\" to use \"$name\"\n";
            }
        }
    }
    
    return $conf->validate_node($node, filename=>$self->{script_base}, node_args=>{no_lists => 1});
}

sub _what_provides {
    my $self = shift;
    my $name = shift;
    my $conf = $self->{conf};
    
    foreach my $module ($conf->modules) {
        return $module, last if
            $conf->is_provided_by_module($module,$name);
    }
}

sub _set_rule {
    my $self = shift;
    my $node = shift;
    my $script_base = $self->{script_base};
    
    my $rule = "";
    my $in_comment = $node->comment;
    foreach my $comment (split "\n", $in_comment) {
        if($comment =~ /^attr[ \t]*NAME[ \t]*=[ \t]*(.*)$/) {
            $rule = $1;
        }
    }
    unless ($rule) {
        $rule = "[UNNAMED RULE]";
        $rule .= " ($script_base line ".$node->test->line.")" if $node->test and $node->test->line;
    }
    $rule = "\"$rule\"";
    $self->{rule} = $rule;
    return;
}

sub _get_rule {
    my $self = shift;
    my $retval = $self->{rule};
    $self->{rule} = "";
    return $retval;
}

1;

