package PureMessage::Policy::Words;

use strict;

use PureMessage::Policy::Message;
use HTML::Parser;

use base 'Exporter';

our @EXPORT_OK = qw(message_words);

sub message_words {
    my ($cb, $msg, %opts) = @_;

    my $max_bytes = delete $opts{max_bytes}; # no limit
    my $chunk_sz  = delete $opts{chunk_size}     || 4096;
    my $per_visit = delete $opts{per_call}       || 1000;
    my $overlap   = delete $opts{overlap}        ||    0; # no overlap
    my $word_re   = delete $opts{word_re}        || qr/\s+/; # whitespace
    die "message_words: unknown option $_" for keys %opts;

    my %ws = map { $_ => 1 } qw(p br body head);
    my (@words, $part, $part_id, $accumulator, $saw_ws);
    $accumulator = '';
    $saw_ws = 0;
    my $html = HTML::Parser->new(
	api_version => 3,
	handlers => {
	    start => [sub {
		my $tag = shift;
		$saw_ws = 1 if $ws{$tag};
	    }, 'tagname'],
	    text => [sub {
		if (length $accumulator) {
		    push @words, $accumulator;
		    if (!$saw_ws and $_[0] =~ s/^(\S+)//) {
			$words[$#words] .= $1;
		    }
		}
		$saw_ws = $_[0] =~ /\s$/;
		push @words, grep length, split($word_re, $_[0]);

		# We need to give the accumulator the last word, in case a
		# word was split in two by the parser.
		$accumulator = @words ? pop @words : '';

		while (@words > $per_visit) {
		    my @w = @words[0 .. $per_visit];
		    &$cb($part, $part_id, \@w);
		    splice(@words, 0, $per_visit - $overlap);
		}
	    }, 'dtext'],
	},
    );
    my $cleanup = sub {
	push @words, split($word_re, $accumulator);
	while (@words > $per_visit) {
	    my @w = @words[0 .. $per_visit];
	    &$cb($part, $part_id, \@w);
	    splice(@words, 0, $per_visit - $overlap);
	}
	&$cb($part, $part_id, \@words) if @words;
	@words = ();
	$accumulator = '';
    };

    $part = $msg;
    $part_id = $msg->part_id;
    if ($msg->header_exists('Subject')) {
	for my $h ($msg->header_decoded('Subject')) {
	    $html->parse($h);
	    $html->eof;
	    &$cleanup;
	}
    }

    my @parts = $msg;
    my $seen = 0;
    while ($part = shift @parts) {
	my @ids = $part->parts;
	unshift @parts, map { $part->part($_) } @ids if @ids;
	$part_id = $part->part_id;
 	my $ct = $part->content_type;
	if ($ct =~ m#^text/(?:plain|html)#i) {
	    my $offset = 0;
	    while (my $len = $part->body(my $buf, $offset, $chunk_sz)) {
		$html->parse($buf);
		$seen   += $len;
		$offset += $len;
	    }

	    # Flush the HTML parser.
	    $html->eof;
	    &$cleanup;
	}
    }
}

1;

=head1 NAME

PureMessage::Policy::Words - extract the words from a message

=head1 SYNOPSIS

   use PureMessage::Policy::Words qw(message_words);

   # Visit every word in every part of a message
   message_words(sub {
       my ($part, $part_id, $words) = @_;
       print "Words in part $part_id: [@$words]\n";
   }, $msg);

   # Visit every word in every part of a message, but the last 5 words will be
   # repeated in the next call.
   message_words(\&visit, $msg, overlap => 5);

   # Only one word per call: (default is 1000)
   message_words(\&visit, $msg, per_call => 1);

=head1 DESCRIPTION

Use C<message_words> to extract the text from a message. You can control how
many words are passed to each invocation of the callback, and how many words
are "overlapped" across multiple invocations.

Words from different message parts are never passed during the same call.

Options:

=over

=item max_bytes

Stop processing after seeing C<max_bytes> bytes of data. Default is 0, or
unlimited.

=item chunk_sz

Read from the message in chunks this big. Default is 4096.

=item per_call

Pass this many words to the callback for each call. Default is 1000.

=item overlap

Repeat this many words between callbacks. Default is 0. Note that overlap
cannot be bigger than per_call.

=item word_re

Pass this regular expression to split() to extract words. Default is C</\s+/>.

=back

=head1 COPYRIGHT

Copyright (C) 2003, ActiveState Corporation.  All Rights Reserved.
